------------------------------------------------------------------------------
-- Chatmanager
-- 
-- Chatmanager is an addon that aims to simplify using chat channels by
-- providing most commonly used (and not so commonly used) features all in one
-- simple window.
-- 
-- Written by Cirk of DoomHammer
------------------------------------------------------------------------------

------------------------------------------------------------------------------
-- AddOn name and version
------------------------------------------------------------------------------
CHATMANAGER_NAME = "Cirk's Chatmanager"
CHATMANAGER_VERSION = "3.2.1"


------------------------------------------------------------------------------
-- Saved state
------------------------------------------------------------------------------
Chatmanager = {};						-- Our global table
ChatmanagerState = {};					-- Will be overridden when loaded, contains for each character:
										--   .Channels = name indexed array of saved channel information
										--   .Options = table of options
										--     .SmartSticky = whether to use smart sticky mode (1 or 0)
										--     .ChatLink = whether to use item linking in private chat channels (1 or 0)
										--     .ReplyLastTell = whether reply goes to last whispered if no more incoming (1 or 0)
										--     .ChatLog = sticky chat logging mode (0 is don't care, 1 is yes but off, 2 is yes and on)
										--     .ShowButton = whether to show menu button or not (1 or 0)
										--     .EnableChatScroll = whether to enable mousewheel scrolling in chat windows (1 or 0)
										--     .HideChatButtons = whether to hide the normal scroll and emote buttons for the chat windows (1 or 0)
										--     .LastStickyType = what sticky type the player was last on (chat type or nil)
										--   .Sticky = table of sticky settings for various chat types
										-- Also contains per server:
										--   .Notes = simple list of player notes, indexed by name


------------------------------------------------------------------------------
-- Globals
------------------------------------------------------------------------------
CHATMANAGER_PLAYER_LIST_BUTTON_HEIGHT = 16;
CHATMANAGER_PLAYER_LIST_BUTTON_COUNT = 17;
CHATMANAGER_LAST_CHANNEL = 10;


------------------------------------------------------------------------------
-- Local defines
------------------------------------------------------------------------------
local CHATMANAGER_PLAYER_UPDATE_INITIAL_DELAY = 5.0;	-- time after starting a list update before removing pending players
local CHATMANAGER_PLAYER_UPDATE_DELAY = 1.0;			-- time after getting an update before removing pending players
local CHATMANAGER_PLAYER_UPDATE_ROSTER_DELAY = 2.0;		-- time after getting a ROSTER update before removing pending players
local CHATMANAGER_PLAYER_REQUEST_DELAY = 2.0;			-- time after cleanup before next request will be issued
local CHATMANAGER_CHANNEL_JOIN_UPDATE_DELAY = 1.5;		-- time after joining channel before process next CHANNEL_UI_UPDATE


------------------------------------------------------------------------------
-- Color information
------------------------------------------------------------------------------
local _colorTextEnabled = {["r"] = 1.0, ["g"] = 0.82, ["b"] = 0};				-- GameFontNormal yellow
local _colorTextDisabled = {["r"] = 0.5, ["g"] = 0.5, ["b"] = 0.5};				-- Light grey
local _highlightAlpha = 0.5;													-- alpha level for highlight texture
local _suspendedAlpha = 0.3;													-- alpha level for suspended channels
local _playerOwner = {["r"] = 1.0, ["g"] = 0.82, ["b"] = 0};
local _playerModerator = {["r"] = 0, ["g"] = 0.82, ["b"] = 1.0};
local _playerMuted = {["r"] = 0.6, ["g"] = 0.6, ["b"] = 0.6};
local _playerNormal = {["r"] = 1.0, ["g"] = 1.0, ["b"] = 1.0};

local _defaultChannelColor = {["r"] = 1, ["g"] = 0.75294122, ["b"] = 0.75294122};


------------------------------------------------------------------------------
-- Local variables
------------------------------------------------------------------------------
local _thisFrame = nil;					-- The frame pointer
local _serverName = nil;				-- set to current realm when loaded
local _playerName = nil;				-- set to current playername when known
local _debugFrame = nil;				-- debug output chat frame
local _functions_post_hooked = nil;		-- flag to indicate when we've done the necessary function hooks

local _channelSavedList = {};			-- list of saved channel information, indexed by lower case channel name (e.g., "mychannel")
										--   .name = channel name in original capitalization (e.g., "MyChannel")
										--   .index = channel index to show to the user for this channel
										--   .password = password for this channel (nil if not used)
										--   .isSticky = sticky status (1 or 0)
										--   .forcePassword = whether to force password or not (1 or 0)
										--   .color = channel color
local _channelList = {};				-- list of channels (indexed by WoW client channel number):
										--   .name = name of channel (short form, e.g., General)
										--   .fullName = name of channel (long form, e.g., General - Ironforge)
										--   .displayIndex = channel index to show to the user for this channel
										--   .channelInfoID = ID corresponding to this channel for GetChannelDisplayInfo
										--   .type = type of channel, where 0 is private, 1 is server, and nil means inactive
										--   .suspended = true if the channel is suspended, nil otherwise
										--   .playerList = list of players (indexed by name) on this channel
										--      .isMuted = true if player is muted
										--      .isModerator = true if player is moderator
										--      .isOwner = true if player is owner
										--      .pending = true if pending delete (nil otherwise)
										--   .playerCount = count of players in .playerList
										--   .filteredList = indexed list of player names in .playerList
										--   .password = password for this channel (nil if not used)
										--   .isSticky = sticky status (1 or 0)
										--   .forcePassword = whether to force password or not (1 or 0)
										--   .amOwner = true if player is owner of the channel
										--   .amModerator = true if player is moderator for this channel
										--   .pending = time that a player list update will expire for this channel
										--   .request = time that a new player list request should be issued for this channel
										--   .userRequest = true if the user has explicitly requested a listing for this channel
										--   .nextUserRequest = true if the next listing for the channel is for the user
local _channelIndexLookupByName = {};			-- client's channel index given the channel name (lower case)
local _channelIndexLookupByDisplayIndex = {};	-- client's channel index given the display index
local _selectedChannelIndex = nil;				-- index in _channelList of currently selected channel (nil means none)
local _selectedChannelDisplayIndex = nil;		-- display index for currently selected channel (nil means none)
local _selectedPlayer = nil;			-- name of player currently targetted

local _convertChannelNumber = nil;		-- set to true when inside ChatEdit_HandleChatType
local _ignoreLinkConversions = nil;		-- set to true when further link conversion should not be done

local _playerListUpdateTimeout = nil;	-- set to time at which the player list should be checked (nil if inactive)
local _channelConnects = nil;			-- list of channels to attempt to reconnect to indexed by name (contains sdata reference)
local _hideChatButtons = nil;			-- whether the chat buttons should be hidden or not currently
local _buttonsHidden = nil;				-- set to 1 if we hide the chat buttons at all during the current session
local _editBoxMoved = nil;				-- set to 1 if we have moved the editbox at all during the current session
local _editBoxFontSized = nil;			-- set to 1 if we have changed the editbox font size at all during the current session

local _channelOriginalText = nil;		-- set to the original channel text if item substitution has been used

local _playerIsOnlineSearchPattern = nil;		-- generated from ERR_FRIEND_ONLINE_SS automatically

local _channelListUpdateTimeout = nil;		-- set to the earliest next time for which the CHANNEL_UI_UPDATE will be handled
local _channelListUpdateIgnore = nil;		-- set to 1 when calling ExpandChannelHeader to avoid processing any CHANNEL_UI_UPDATE events
local _channelListUpdateInProgress = nil;	-- set to 1 when Chatmanager is executing any of its channel list hooked functions to avoid infinite loops
local _channelListButtonLookupById = {};	-- table allowing fast conversion from channel ID to button index

local _hasTargetExact = nil;			-- Set to true if we know there is a /targetexact macro (WoW 2.3 onwards)


------------------------------------------------------------------------------
-- Hooked function originals
------------------------------------------------------------------------------
local _original_GetChannelName = nil;
local _original_GetChannelList = nil;
local _original_ChatFrame_ConfigEventHandler = nil;
local _original_ChatFrame_MessageEventHandler = nil;
local _original_ChatEdit_UpdateHeader = nil;
local _original_JoinPermanentChannel = nil;
local _original_ListChannelByName = nil;
local _original_ChatEdit_HandleChatType = nil;
local _original_ChatEdit_OnEscapePressed = nil;
local _original_ChatEdit_AddHistory = nil;
local _original_SendChatMessage = nil;
local _original_FCF_SetButtonSide = nil;
local _original_SetChannelPassword = nil;

local _original_SlashCommand_LEAVE = nil;
local _original_SlashCommand_LIST_CHANNEL = nil;
local _original_SlashCommand_CHAT_PASSWORD = nil;
local _original_SlashCommand_CHAT_OWNER = nil;
local _original_SlashCommand_CHAT_MODERATOR = nil;
local _original_SlashCommand_CHAT_UNMODERATOR = nil;
local _original_SlashCommand_CHAT_MODERATE = nil;
local _original_SlashCommand_CHAT_MUTE = nil;
local _original_SlashCommand_CHAT_UNMUTE = nil;
local _original_SlashCommand_CHAT_CINVITE = nil;
local _original_SlashCommand_CHAT_KICK = nil;
local _original_SlashCommand_CHAT_BAN = nil;
local _original_SlashCommand_CHAT_UNBAN = nil;
local _original_SlashCommand_CHAT_ANNOUNCE = nil;

local _original_ChannelList_UpdateVoice = nil;


------------------------------------------------------------------------------
-- Hooked function originals for other addons
------------------------------------------------------------------------------
local _original_FLT_ParseChatMessage = nil;			-- Linkerator


------------------------------------------------------------------------------
-- Popup defines (see Blizzard's StaticPopup.lua)
------------------------------------------------------------------------------
local _chatmanagerConfirmLeavePopup = {
	text = CHATMANAGER_TEXT.CONFIRM_LEAVE_FORMAT,
	button1 = TEXT(YES),
	button2 = TEXT(NO),
	sound = "igCharacterInfoOpen",
	OnAccept = function(self, data)
		Chatmanager.HandlePopupAccept("LEAVE", data);
	end,
	timeout = 60,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1,
	showAlert = 1,
};


local _chatmanagerEnterChannelNamePopup = {
	text = CHATMANAGER_TEXT.ENTER_NEW_CHANNEL_TEXT,
	button1 = TEXT(ACCEPT),
	button2 = TEXT(CANCEL),
	hasEditBox = 1,
	hasWideEditBox = 1,
	maxLetters = 60,
	OnShow = function(self)
		local editBox = getglobal(self:GetName().."WideEditBox");
		editBox:SetText("");
		editBox:HighlightText();
		editBox:SetFocus();
		self:SetWidth(420);
		getglobal(self:GetName().."Text"):SetWidth(390);
	end,
	OnHide = function(self)
		if (ChatFrameEditBox:IsShown()) then
			ChatFrameEditBox:SetFocus();
		end
		getglobal(self:GetName().."WideEditBox"):SetText("");
		self:SetWidth(320);
		getglobal(self:GetName().."Text"):SetWidth(290);
	end,
	OnAccept = function(self, data)
		local editBox = getglobal(self:GetName().."WideEditBox");
		Chatmanager.HandlePopupAccept("JOIN", data, editBox:GetText());
	end,
	EditBoxOnEnterPressed = function(self, data)
		Chatmanager.HandlePopupAccept("JOIN", data, self:GetText());
		self:GetParent():Hide();
	end,
	EditBoxOnEscapePressed = function(self)
		self:GetParent():Hide();
	end,
	timeout = 60,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1,
};


local _chatmanagerEnterPlayerNotePopup = {
	text = CHATMANAGER_TEXT.ENTER_PLAYER_NOTE_TEXT,
	button1 = TEXT(ACCEPT),
	button2 = TEXT(CANCEL),
	hasEditBox = 1,
	hasWideEditBox = 1,
	maxLetters = 60,
	OnShow = function(self)
		local editBox = getglobal(self:GetName().."WideEditBox");
		editBox:SetText(Chatmanager.GetPopupData("PLAYERNOTE"));
		editBox:HighlightText();
		editBox:SetFocus();
		self:SetWidth(420);
		getglobal(self:GetName().."Text"):SetWidth(390);
	end,
	OnHide = function(self)
		if (ChatFrameEditBox:IsShown()) then
			ChatFrameEditBox:SetFocus();
		end
		getglobal(self:GetName().."WideEditBox"):SetText("");
		self:SetWidth(320);
		getglobal(self:GetName().."Text"):SetWidth(290);
	end,
	OnAccept = function(self, data)
		local editBox = getglobal(self:GetName().."WideEditBox");
		Chatmanager.HandlePopupAccept("PLAYERNOTE", data, editBox:GetText());
	end,
	EditBoxOnEnterPressed = function(self, data)
		Chatmanager.HandlePopupAccept("PLAYERNOTE", data, self:GetText());
		self:GetParent():Hide();
	end,
	EditBoxOnEscapePressed = function(self)
		self:GetParent():Hide();
	end,
	timeout = 60,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1,
};


------------------------------------------------------------------------------
-- Setup functions
------------------------------------------------------------------------------
function Chatmanager.HookFunctions()
	-- Note that although Chatmanager uses a lot of function hooks, most of
	-- these are lightweight wrappers or near-identical replacements, and are
	-- intended to maintain the illusion that the channel order selected by
	-- the user is being used by the client, rather than just simulated by the
	-- addon.
	if (not _original_GetChannelName) then
		_original_GetChannelName = GetChannelName;
		GetChannelName = Chatmanager.GetChannelName;
	end
	if (not _original_GetChannelList) then
		_original_GetChannelList = GetChannelList;
		GetChannelList = Chatmanager.GetChannelList;
	end
	if (not _original_ChatFrame_ConfigEventHandler) then
		_original_ChatFrame_ConfigEventHandler = ChatFrame_ConfigEventHandler;
		ChatFrame_ConfigEventHandler = Chatmanager.ChatFrame_ConfigEventHandler;
	end
	if (not _original_ChatFrame_MessageEventHandler) then
		_original_ChatFrame_MessageEventHandler = ChatFrame_MessageEventHandler;
		ChatFrame_MessageEventHandler = Chatmanager.ChatFrame_MessageEventHandler;
	end
	if (not _original_ChatEdit_UpdateHeader) then
		_original_ChatEdit_UpdateHeader = ChatEdit_UpdateHeader;
		ChatEdit_UpdateHeader = Chatmanager.ChatEdit_UpdateHeader;
	end
	if (not _original_JoinPermanentChannel) then
		_original_JoinPermanentChannel = JoinPermanentChannel;
		JoinPermanentChannel = Chatmanager.JoinPermanentChannel;
	end
	if (not _original_ListChannelByName) then
		_original_ListChannelByName = ListChannelByName;
		ListChannelByName = Chatmanager.ListChannelByName;
	end
	if (not _original_ChatEdit_HandleChatType) then
		_original_ChatEdit_HandleChatType = ChatEdit_HandleChatType;
		ChatEdit_HandleChatType = Chatmanager.ChatEdit_HandleChatType;
	end
	if (not _original_ChatEdit_OnEscapePressed) then
		_original_ChatEdit_OnEscapePressed = ChatEdit_OnEscapePressed;
		ChatEdit_OnEscapePressed = Chatmanager.ChatEdit_OnEscapePressed;
	end
	if (not _original_ChatEdit_AddHistory) then
		_original_ChatEdit_AddHistory = ChatEdit_AddHistory;
		ChatEdit_AddHistory = Chatmanager.ChatEdit_AddHistory;
	end
	if (not _original_SendChatMessage) then
		_original_SendChatMessage = SendChatMessage;
		SendChatMessage = Chatmanager.SendChatMessage;
	end
	if (not _original_FCF_SetButtonSide) then
		_original_FCF_SetButtonSide = FCF_SetButtonSide;
		FCF_SetButtonSide = Chatmanager.FCF_SetButtonSide;
	end
	if (not _original_SetChannelPassword) then
		_original_SetChannelPassword = SetChannelPassword;
		SetChannelPassword = Chatmanager.SetChannelPassword;
	end

	-- Slash command function hooks
	if (not _original_SlashCommand_LEAVE) then
		_original_SlashCommand_LEAVE = SlashCmdList["LEAVE"];
		SlashCmdList["LEAVE"] = Chatmanager.SlashCommand_LEAVE;
	end
	if (not _original_SlashCommand_LIST_CHANNEL) then
		_original_SlashCommand_LIST_CHANNEL = SlashCmdList["LIST_CHANNEL"];
		SlashCmdList["LIST_CHANNEL"] = Chatmanager.SlashCommand_LIST_CHANNEL;
	end
	if (not _original_SlashCommand_CHAT_PASSWORD) then
		_original_SlashCommand_CHAT_PASSWORD = SlashCmdList["CHAT_PASSWORD"];
		SlashCmdList["CHAT_PASSWORD"] = Chatmanager.SlashCommand_CHAT_PASSWORD;
	end
	if (not _original_SlashCommand_CHAT_OWNER) then
		_original_SlashCommand_ = SlashCmdList["CHAT_OWNER"];
		SlashCmdList["CHAT_OWNER"] = Chatmanager.SlashCommand_CHAT_OWNER;
	end
	if (not _original_SlashCommand_CHAT_MODERATOR) then
		_original_SlashCommand_CHAT_MODERATOR = SlashCmdList["CHAT_MODERATOR"];
		SlashCmdList["CHAT_MODERATOR"] = Chatmanager.SlashCommand_CHAT_MODERATOR;
	end
	if (not _original_SlashCommand_CHAT_UNMODERATOR) then
		_original_SlashCommand_CHAT_UNMODERATOR = SlashCmdList["CHAT_UNMODERATOR"];
		SlashCmdList["CHAT_UNMODERATOR"] = Chatmanager.SlashCommand_CHAT_UNMODERATOR;
	end
	if (not _original_SlashCommand_CHAT_MODERATE) then
		_original_SlashCommand_CHAT_MODERATE = SlashCmdList["CHAT_MODERATE"];
		SlashCmdList["CHAT_MODERATE"] = Chatmanager.SlashCommand_CHAT_MODERATE;
	end
	if (not _original_SlashCommand_CHAT_MUTE) then
		_original_SlashCommand_CHAT_MUTE = SlashCmdList["CHAT_MUTE"];
		SlashCmdList["CHAT_MUTE"] = Chatmanager.SlashCommand_CHAT_MUTE;
	end
	if (not _original_SlashCommand_CHAT_UNMUTE) then
		_original_SlashCommand_CHAT_UNMUTE = SlashCmdList["CHAT_UNMUTE"];
		SlashCmdList["CHAT_UNMUTE"] = Chatmanager.SlashCommand_CHAT_UNMUTE;
	end
	if (not _original_SlashCommand_CHAT_CINVITE) then
		_original_SlashCommand_CHAT_CINVITE = SlashCmdList["CHAT_CINVITE"];
		SlashCmdList["CHAT_CINVITE"] = Chatmanager.SlashCommand_CHAT_CINVITE;
	end
	if (not _original_SlashCommand_CHAT_KICK) then
		_original_SlashCommand_CHAT_KICK = SlashCmdList["CHAT_KICK"];
		SlashCmdList["CHAT_KICK"] = Chatmanager.SlashCommand_CHAT_KICK;
	end
	if (not _original_SlashCommand_CHAT_BAN) then
		_original_SlashCommand_CHAT_BAN = SlashCmdList["CHAT_BAN"];
		SlashCmdList["CHAT_BAN"] = Chatmanager.SlashCommand_CHAT_BAN;
	end
	if (not _original_SlashCommand_CHAT_UNBAN) then
		_original_SlashCommand_CHAT_UNBAN = SlashCmdList["CHAT_UNBAN"];
		SlashCmdList["CHAT_UNBAN"] = Chatmanager.SlashCommand_CHAT_UNBAN;
	end
	if (not _original_SlashCommand_CHAT_ANNOUNCE) then
		_original_SlashCommand_CHAT_ANNOUNCE = SlashCmdList["CHAT_ANNOUNCE"];
		SlashCmdList["CHAT_ANNOUNCE"] = Chatmanager.SlashCommand_CHAT_ANNOUNCE;
	end

	-- Of the WoW 2.2 channel menu functions, we only need to intercept calls
	-- to ChannelList_UpdateVoice due to interaction between when the
	-- CHANNEL_VOICE_UPDATE event occurs and when GetChannelDisplayInfo will
	-- present the right information for the affected channel.
	if (not _original_ChannelList_UpdateVoice) then
		_original_ChannelList_UpdateVoice = ChannelList_UpdateVoice;
		ChannelList_UpdateVoice = Chatmanager.ChannelList_UpdateVoice;
	end
	
	-- With the 2.0 client we can no longer directly hook some functions (such
	-- as ChatEdit_SendText) and keep the ability to issue "secure" commands
	-- within the hooked context, so instead we post-hook the functions (via
	-- hooksecurefunc) instead, where our "tainted" code is then called after
	-- the original function has completed.
	if (not _functions_post_hooked) then
		hooksecurefunc("ChatEdit_SendText", Chatmanager.ChatEdit_SendText);
		hooksecurefunc("UnitPopup_ShowMenu", Chatmanager.UnitPopup_ShowMenu);
		hooksecurefunc("ChatFrame_OpenChat", Chatmanager.ChatFrame_OpenChat);
		hooksecurefunc("FCF_SetChatWindowFontSize", Chatmanager.FCF_SetChatWindowFontSize);
		hooksecurefunc("ChannelList_Update", Chatmanager.ChannelList_Update);
		hooksecurefunc("ChannelList_CountUpdate", Chatmanager.ChannelList_CountUpdate);
		hooksecurefunc("ChannelList_UpdateHighlight", Chatmanager.ChannelList_UpdateHighlight);
		hooksecurefunc("CreateChatChannelList", Chatmanager.CreateChatChannelList);
		_functions_post_hooked = true;
	end

	-- Handle other addon interaction problems
	if (LinkeratorFrame and FLT_ParseChatMessage) then
		-- For the Fizzwidget Linkerator (http://fizzwidget.com/linkerator/)
		if (not _original_FLT_ParseChatMessage) then
			_original_FLT_ParseChatMessage = FLT_ParseChatMessage;
			FLT_ParseChatMessage = Chatmanager.FLT_ParseChatMessage;
		end
	end
end


function Chatmanager.RegisterEvents()
	_thisFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
	_thisFrame:RegisterEvent("CHANNEL_UI_UPDATE");
	_thisFrame:RegisterEvent("CHANNEL_ROSTER_UPDATE");
	_thisFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
	_thisFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
	_thisFrame:RegisterEvent("CHANNEL_PASSWORD_REQUEST");
end


function Chatmanager.UnregisterEvents()
	_thisFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
	_thisFrame:RegisterEvent("CHANNEL_UI_UPDATE");
	_thisFrame:RegisterEvent("CHANNEL_ROSTER_UPDATE");
	_thisFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
	_thisFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
	_thisFrame:UnregisterEvent("CHANNEL_PASSWORD_REQUEST");
end


function Chatmanager.CheckLoadedVariables()
	-- Make sure that all loaded variables have valid values and performs any
	-- fixups required for version to version changes of Chatmanager or for
	-- saved versus active values.

	-- Convert from old per-character option data
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Options) then
		if (ChatmanagerState.Options.Default) then
			ChatmanagerState.Options.SmartSticky = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.SmartSticky;
			ChatmanagerState.Options.ChatLink = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.ChatLink;
			ChatmanagerState.Options.ReplyLastTell = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.ReplyLastTell;
			ChatmanagerState.Options.ChatLog = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.ChatLog;
			ChatmanagerState.Options.ShowButton = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.ShowButton;
			ChatmanagerState.Options.EnableChatScroll = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.EnableChatScroll;
			ChatmanagerState.Options.HideChatButtons = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.HideChatButtons;
			ChatmanagerState.Options.EditBoxAtTop = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.EditBoxAtTop;
			ChatmanagerState.Options.EditBoxNoAlt = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.EditBoxNoAlt;
			ChatmanagerState.Options.EditBoxFontSize = ChatmanagerState.Servers[_serverName].Characters[_playerName].Options.EditBoxFontSize;
		end
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Options = nil;
	end
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].ShowPlayerNote) then
		if (ChatmanagerState.Options.Default) then
			ChatmanagerState.Options.ShowPlayerNote = ChatmanagerState.Servers[_serverName].Characters[_playerName].ShowPlayerNote;
		end
		ChatmanagerState.Servers[_serverName].Characters[_playerName].ShowPlayerNote = nil;
	end
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].UseGuildNote) then
		if (ChatmanagerState.Options.Default) then
			ChatmanagerState.Options.UseGuildNote = ChatmanagerState.Servers[_serverName].Characters[_playerName].UseGuildNote;
		end
		ChatmanagerState.Servers[_serverName].Characters[_playerName].UseGuildNote = nil;
	end
	ChatmanagerState.Options.Default = nil;
	ChatmanagerState.Options.ChatLinkFormat = nil;

	if (not ChatmanagerState.Options.ShowPlayerNote) then
		ChatmanagerState.Options.ShowPlayerNote = 1;
	end

	if (ChatmanagerState.Options.ChatLog == 2) then
		ChatmanagerState.Options.ChatLog = 1;
		if (not LoggingChat()) then
			LoggingChat(true);
		end
		local info = ChatTypeInfo["SYSTEM"];
		DEFAULT_CHAT_FRAME:AddMessage(TEXT(CHATLOGENABLED), info.r, info.g, info.b, info.id);
	end

	local savedChannels = ChatmanagerState.Servers[_serverName].Characters[_playerName].Channels;
	local convert = nil;
	for name, sdata in pairs(savedChannels) do
		local lname = string.lower(name);
		if ((name ~= lname) or not sdata.name) then
			convert = true;
			break;
		end
	end
	if (convert) then
		local newChannels = {};
		for name, sdata in pairs(savedChannels) do
			local lname = string.lower(name);
			if (not sdata.name) then
				sdata.name = name;
			end
			newChannels[lname] = sdata;
		end
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Channels = newChannels;
	end
end


function Chatmanager.Initialize()
	-- Called once the player's information has been loaded, this function
	-- fills in the current channel list and then loads the saved channel list
	-- information from ChatmanagerState so it can make changes for the actual
	-- versus displayed channel index.  Once this is done, the reference
	-- tables are filled in for speed of access later.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Initialize()");
	end

	-- First get list of current channels from the client using the new
	-- GetChannelDisplayInfo function (this will normally be
	-- empty unless we have done a console reload).
	_channelList = {};
	_channelIndexLookupByName = {};
	for index = 1, CHATMANAGER_LAST_CHANNEL do
		_channelList[index] = {};
	end
	for id = 1, MAX_CHANNEL_BUTTONS do
		local name, header, collapsed, channelNumber, count, active, category = GetChannelDisplayInfo(id);
		if (header) then
			if (collapsed) then
				-- We need all the tabs expanded in order to read the channel
				-- information inside of it (this happens immediately).
				-- Expanding the channel header will trigger a
				-- CHANNEL_UI_UPDATE event which will be processed immediately
				-- if we are registered for it, which we want to ignore.
				_channelListUpdateIgnore = 1;
				ExpandChannelHeader(id);
				_channelListUpdateIgnore = nil;
			end
		elseif (name and channelNumber and (category ~= "CHANNEL_CATEGORY_GROUP")) then
			local index = tonumber(channelNumber);
			local cdata = _channelList[index or 0];
			if (cdata) then
				local _, fullName = _original_GetChannelName(tostring(index));
				cdata.name = name;
				cdata.fullName = fullName;
				_channelIndexLookupByName[string.lower(name)] = index;
				if (category == "CHANNEL_CATEGORY_WORLD") then
					cdata.type = 1;
					if (not active) then
						cdata.suspended = true;
					end
				else
					cdata.type = 0;
				end
				cdata.channelInfoID = id;
			end
		elseif (id > GetNumDisplayChannels()) then
			break;
		end
	end

	-- Now use the saved list to update the display indices and other saved
	-- parameters
	_channelSavedList = ChatmanagerState.Servers[_serverName].Characters[_playerName].Channels;
	_channelIndexLookupByDisplayIndex = {};
	for i = 1, CHATMANAGER_LAST_CHANNEL do
		if (_channelList[i].name) then
			local lname = string.lower(_channelList[i].name);
			local sdata = _channelSavedList[lname];
			if (sdata) then
				_channelList[i].password = sdata.password;
				_channelList[i].isSticky = sdata.isSticky;
				_channelList[i].forcePassword = sdata.forcePassword;
				if (sdata.index and not _channelIndexLookupByDisplayIndex[sdata.index]) then
					_channelList[i].displayIndex = sdata.index;
					_channelIndexLookupByDisplayIndex[sdata.index] = i;
				end
				if (sdata.color) then
					ChangeChatColor("CHANNEL"..i, sdata.color.r, sdata.color.g, sdata.color.b);
				end
			end
		end
	end

	-- Find any remaining entries that don't have a displayIndex assigned yet
	-- and give them one.
	for i = 1, CHATMANAGER_LAST_CHANNEL do
		if (_channelList[i].name and not _channelList[i].displayIndex) then
			for j = 1, CHATMANAGER_LAST_CHANNEL do
				if (not _channelIndexLookupByDisplayIndex[j]) then
					_channelList[i].displayIndex = j;
					_channelIndexLookupByDisplayIndex[j] = i;
					break;
				end
			end
		end
	end

	-- Clean up WoW's Chatframe event registration to reduce overheads a bit,
	-- and make sure the first tab is selected by default.
	tinsert(ChatTypeGroup["CHANNEL"], "CHAT_MSG_CHANNEL");
	for tab = 1, NUM_CHAT_WINDOWS do
		local chatFrame = getglobal("ChatFrame"..tab);
		local _, _, _, _, _, _, shown, _, docked = GetChatWindowInfo(tab);
		if (shown or docked) then
			chatFrame:RegisterEvent("CHAT_MSG_CHANNEL");
		else
			chatFrame:UnregisterEvent("CHAT_MSG_CHANNEL");
		end
	end
	if (ChatFrame1.isDocked) then
		FCF_SelectDockFrame(ChatFrame1);
	end

	-- If GetSelectedDisplayChannel is set, it probably means we did a reload
	-- of the UI.  In this circumstance we need to force a (temporary) change
	-- in the displayed channel or else we won't get meaningful data from it.
	local currentChannel = GetSelectedDisplayChannel();
	if (currentChannel) then
		for index, cdata in pairs(_channelList) do
			if (cdata and cdata.name and cdata.channelInfoID and not cdata.suspended) then
				if (cdata.channelInfoID ~= currentChannel) then
					SetSelectedDisplayChannel(cdata.channelInfoID);
				end
			end
		end
		SetSelectedDisplayChannel(currentChannel);
	end

	-- Lastly request the player list for every non-server channel.
	for index, cdata in ipairs(_channelList) do
		if (cdata.name and (cdata.type == 0)) then
			Chatmanager.RequestPlayerList(index, false, false);
		end
	end
end


function Chatmanager.UpdateSticky()
	for type, value in pairs(ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky) do
		ChatTypeInfo[type].sticky = value;
	end
	local lastSticky = ChatmanagerState.Servers[_serverName].Characters[_playerName].LastStickyType;
	if (lastSticky and ChatTypeInfo[lastSticky].sticky == 1) then
		if ((lastSticky == "RAID") and (GetNumRaidMembers() > 0)) then
			ChatFrameEditBox:SetAttribute("chatType", "RAID");
			ChatFrameEditBox:SetAttribute("stickyType", "RAID");
		elseif ((lastSticky == "BATTLEGROUND") and (GetNumRaidMembers() > 0)) then
			ChatFrameEditBox:SetAttribute("chatType", "BATTLEGROUND");
			ChatFrameEditBox:SetAttribute("stickyType", "BATTLEGROUND");
		elseif ((lastSticky == "PARTY") and (GetNumPartyMembers() > 0)) then
			ChatFrameEditBox:SetAttribute("chatType", "PARTY");
			ChatFrameEditBox:SetAttribute("stickyType", "PARTY");
		elseif ((ChatTypeInfo["GUILD"].sticky == 1) and IsInGuild()) then
			ChatFrameEditBox:SetAttribute("chatType", "GUILD");
			ChatFrameEditBox:SetAttribute("stickyType", "GUILD");
		end
	elseif ((ChatTypeInfo["GUILD"].sticky == 1) and IsInGuild()) then
		ChatFrameEditBox:SetAttribute("chatType", "GUILD");
		ChatFrameEditBox:SetAttribute("stickyType", "GUILD");
	end
end


function Chatmanager.UpdateScrollingAndButtons()
	-- Enable or disable mouse wheel events, and show or hide the scroll
	-- buttons for each window as needed.  To keep the buttons hidden, a new
	-- field is added to each frame object to contain the original OnShow
	-- handler, and the OnShow handler replaced with our own version.
	-- Similarly a new field is added to the bottom button for each frame.
	-- In order to prevent "touching" the chat frame button states when we are
	-- not controlling them (but which other components might be) we use the
	-- _buttonsHidden flag to determine if we have actively hidden the buttons
	-- in this session, and therefore whether it is valid to unhide them if
	-- such a action is required.
	local enable = ChatmanagerState.Options.EnableChatScroll;
	local hidden = ChatmanagerState.Options.HideChatButtons;
	if ((enable == 1) and (hidden == 1)) then
		_hideChatButtons = 1;
		_buttonsHidden = 1;
	else
		_hideChatButtons = 0;
	end
	for i = 1, 7 do
		local frame = getglobal("ChatFrame"..i);
		local bottomButton = getglobal("ChatFrame"..i.."BottomButton");
		if (frame) then
			if (enable == 1) then
				frame:EnableMouseWheel(1);
				frame.chatmanagerOrigOnMouseWheel = frame:GetScript("OnMouseWheel");
				frame:SetScript("OnMouseWheel", Chatmanager.ChatFrameOnScroll);
			elseif (frame:GetScript("OnMouseWheel") == Chatmanager.ChatFrameOnScroll) then
				frame:EnableMouseWheel(0);
				frame:SetScript("OnMouseWheel", frame.chatmanagerOrigOnMouseWheel);
			end
			if (_hideChatButtons == 1) then
				if (not frame.chatmanagerOrigOnShow) then
					frame.chatmanagerOrigOnShow = frame:GetScript("OnShow");
					frame:SetScript("OnShow", Chatmanager.ChatFrameOnShow);
					if (frame:IsShown()) then
						frame:Hide();
						frame:Show();
					end
				end
			elseif (frame.chatmanagerOrigOnShow) then
				frame:SetScript("OnShow", frame.chatmanagerOrigOnShow);
				if (frame:IsShown()) then
					frame:Hide();
					frame:Show();
				end
				frame.chatmanagerOrigOnShow = nil;
			end
		end
		if (bottomButton) then
			if (_hideChatButtons == 1) then
				if (not bottomButton.chatmanagerOrigOnClick) then
					bottomButton.chatmanagerOrigOnClick = bottomButton:GetScript("OnClick");
					bottomButton:SetScript("OnClick", Chatmanager.ChatFrameBottomButtonOnClick);
				end
				bottomButton:ClearAllPoints();
				bottomButton:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, -4);
			elseif (_buttonsHidden == 1) then
				bottomButton:ClearAllPoints();
				if (FCF_GetButtonSide(frame) == "left") then
					bottomButton:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", -32, -4);
				else
					bottomButton:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 0, -4);
				end
				if (bottomButton.chatmanagerOrigOnClick) then
					bottomButton:SetScript("OnClick", bottomButton.chatmanagerOrigOnClick);
					bottomButton.chatmanagerOrigOnClick = nil;
				end
			end
		end
	end
	if (_hideChatButtons == 1) then
		ChatmanagerMenuButtonFrameMenuButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
		ChatmanagerMenuButtonFrameMenuButton.tooltipText3 = CHATMANAGER_TEXT.MENU_BUTTON_MENU_TOOLTIP;
		ChatFrameMenuButton:Hide();
		ChatMenu:SetMovable(true);
		ChatMenu:SetClampedToScreen(true);
		ChatMenu:ClearAllPoints();
		ChatMenu:SetPoint("BOTTOMLEFT", "ChatmanagerMenuButtonFrame", "TOPRIGHT");
		ChatMenu:SetUserPlaced(false);
	else
		ChatmanagerMenuButtonFrameMenuButton:RegisterForClicks("LeftButtonUp");
		ChatmanagerMenuButtonFrameMenuButton.tooltipText3 = nil;
		if (_buttonsHidden == 1) then
			ChatFrameMenuButton:Show();
			ChatMenu:ClearAllPoints();
			ChatMenu:SetPoint("BOTTOMLEFT", "ChatFrameMenuButton", "TOPRIGHT");
			ChatMenu:SetUserPlaced(false);
		end
	end
	if (not ChatmanagerMenuButtonFrame:IsUserPlaced()) then
		-- Change the position of the menu button depending on the hidden or
		-- shown status of the chat scroll buttons
		ChatmanagerMenuButtonFrame:ClearAllPoints();
		if (_hideChatButtons == 1) then
			ChatmanagerMenuButtonFrame:SetPoint("TOPRIGHT", "ChatFrame1", "TOPRIGHT");
		else
			ChatmanagerMenuButtonFrame:SetPoint("BOTTOM", "ChatFrameMenuButton", "TOP");
		end
	end
end


function Chatmanager.UpdateEditBoxPosition()
	-- Updates the position of the editbox.  To avoid touching the editbox if
	-- the user has not elected this option, the _editBoxMoved flag is used to
	-- keep track of whether we have "touched" it or not.
	local atTop = ChatmanagerState.Options.EditBoxAtTop;
	if (atTop == 1) then
		ChatFrameEditBox:ClearAllPoints();
		ChatFrameEditBox:SetPoint("TOPLEFT", "ChatFrame1", "TOPLEFT", -5, 6);
		ChatFrameEditBox:SetPoint("TOPRIGHT", "ChatFrame1", "TOPRIGHT", 5, 6);
		_editBoxMoved = 1;
	elseif (_editBoxMoved) then
		ChatFrameEditBox:ClearAllPoints();
		ChatFrameEditBox:SetPoint("TOPLEFT", "ChatFrame1", "BOTTOMLEFT", -5, -2);
		ChatFrameEditBox:SetPoint("TOPRIGHT", "ChatFrame1", "TOPRIGHT", 5, -2);
	end
end


function Chatmanager.UpdateEditBoxFont(fontSize)
	-- Sets the editBox font size (if enabled) or restores it otherwise.  To
	-- avoid touching the editbox if the user has not elected this option, the
	-- _editBoxFontSized flag is used to keep track of whether we have
	-- "touched" it or not.
	local setFont = ChatmanagerState.Options.EditBoxFontSize;
	if (setFont == 1) then
		if (not fontSize) then
			local _, fontHeight, _ = ChatFrame1:GetFont();
			fontSize = fontHeight;
		end
		local fontFile, unused, fontFlags = ChatFrameEditBox:GetFont();
		ChatFrameEditBox:SetFont(fontFile, fontSize, fontFlags);
		fontFile, unused, fontFlags = ChatFrameEditBoxHeader:GetFont();
		ChatFrameEditBoxHeader:SetFont(fontFile, fontSize, fontFlags);
		_editBoxFontSized = 1;
	elseif (_editBoxFontSized) then
		local fontFile, unused, fontFlags = ChatFrameEditBox:GetFont();
		ChatFrameEditBox:SetFont(fontFile, 14, fontFlags);
		fontFile, unused, fontFlags = ChatFrameEditBoxHeader:GetFont();
		ChatFrameEditBoxHeader:SetFont(fontFile, 14, fontFlags);
		_editBoxFontSized = 1;
	end
end


function Chatmanager.UpdateEditBoxAttribs()
	-- Sets miscellaneous attributes of the editbox
	ChatFrameEditBox:SetAltArrowKeyMode(ChatmanagerState.Options.EditBoxNoAlt ~= 1);
end


function Chatmanager.VariablesLoaded()
	if (not ChatmanagerState) then
		ChatmanagerState = {};
	end
	if (not ChatmanagerState.Servers) then
		ChatmanagerState.Servers = {};
	end
	if (not ChatmanagerState.Servers[_serverName]) then
		ChatmanagerState.Servers[_serverName] = {};
	end
	if (not ChatmanagerState.Servers[_serverName].Characters) then
		ChatmanagerState.Servers[_serverName].Characters = {};
	end
	if (not ChatmanagerState.Servers[_serverName].Notes) then
		ChatmanagerState.Servers[_serverName].Notes = {};
	end
end


function Chatmanager.PlayerLogin()
	-- Make sure basic saved variable structure is present
	if (not ChatmanagerState.Servers[_serverName].Characters[_playerName]) then
		ChatmanagerState.Servers[_serverName].Characters[_playerName] = {};
	end
	if (not ChatmanagerState.Servers[_serverName].Characters[_playerName].Channels) then
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Channels = {};
	end
	if (not ChatmanagerState.Options) then
		ChatmanagerState.Options = {};
		ChatmanagerState.Options.SmartSticky = 0;
		ChatmanagerState.Options.ChatLink = 0;
		ChatmanagerState.Options.ReplyLastTell = 0;
		ChatmanagerState.Options.ChatLog = 0;
		ChatmanagerState.Options.ShowButton = 1;
		ChatmanagerState.Options.EnableChatScroll = 0;
		ChatmanagerState.Options.HideChatButtons = 0;
		ChatmanagerState.Options.EditBoxAtTop = 0;
		ChatmanagerState.Options.EditBoxNoAlt = 0;
		ChatmanagerState.Options.EditBoxFontSize = 0;
		ChatmanagerState.Options.ShowPlayerNote = 1;
		ChatmanagerState.Options.UseGuildNote = 1;
		ChatmanagerState.Options.Default = 1;
	end
	if (not ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky) then
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky = {};
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["SAY"] = ChatTypeInfo["SAY"].sticky;
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["PARTY"] = ChatTypeInfo["PARTY"].sticky;
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["RAID"] = ChatTypeInfo["RAID"].sticky;
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["GUILD"] = ChatTypeInfo["GUILD"].sticky;
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["OFFICER"] = ChatTypeInfo["OFFICER"].sticky;
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["WHISPER"] = ChatTypeInfo["WHISPER"].sticky;
	end
	if (not ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["BATTLEGROUND"]) then
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky["BATTLEGROUND"] = ChatTypeInfo["BATTLEGROUND"].sticky;
	end

	-- Process the loaded variables (and do past version cleanups)
	Chatmanager.CheckLoadedVariables();

	-- Hook required functions
	Chatmanager.HookFunctions();

	-- Begin processing of the current channels (if any)
	Chatmanager.Initialize();

	-- Update sticky mode for non-channels
	Chatmanager.UpdateSticky();

	-- Update chat scrolling and button shown/hidden states, and editbox position/status
	Chatmanager.UpdateScrollingAndButtons();
	Chatmanager.UpdateEditBoxPosition();
	Chatmanager.UpdateEditBoxFont();
	Chatmanager.UpdateEditBoxAttribs();

	-- Show the Chatmanager button (if enabled)
	if (ChatmanagerState.Options.ShowButton == 1) then
		ChatmanagerMenuButtonFrame:Show();
	end

	-- Determine whether /targetexact exists or not for use in our macros
	if (SLASH_TARGET_EXACT1 and IsSecureCmd(SLASH_TARGET_EXACT1)) then
		_hasTargetExact = 1;
	end
end


function Chatmanager.PlayerLogout()
	-- Just build and save the new _channelSavedList.  Note that in regards to
	-- password protected channels and the client's not saving the password,
	-- some Addon developers recommend simply leaving the channels on logout.
	-- However this doesn't appear to work properly: if we are doing a console
	-- reload then the YOU_LEFT message arrives after the reload and causes
	-- some sequencing issues; and a proper logout doesn't record that you
	-- have actually left the channel, and so logs back in with the default
	-- empty password anyway.  Since here we don't have to worry about the
	-- order in which channels are joined explicitly, we will deal with the
	-- password issue by reattempting at login instead.

	-- First copy relevant details from current channel list
	_channelSavedList = {};
	for i = 1, CHATMANAGER_LAST_CHANNEL do
		if (_channelList[i].name) then
			sdata = {};
			sdata.index = _channelList[i].displayIndex;
			if (_channelList[i].password and (_channelList[i].password ~= "") and (_channelList[i].type == 0)) then
				sdata.password = _channelList[i].password;
			end
			sdata.isSticky = _channelList[i].isSticky;
			sdata.forcePassword = _channelList[i].forcePassword;
			local info = ChatTypeInfo["CHANNEL"..i];
			sdata.color = {};
			sdata.color.r = info.r;
			sdata.color.g = info.g;
			sdata.color.b = info.b;
			sdata.name = _channelList[i].name;
			_channelSavedList[string.lower(_channelList[i].name)] = sdata;
		end
	end
	-- For all password protected channels, build a table of which tabs each
	-- channel is assigned to for later restoration.
	local windowList;
	for tab = 1, NUM_CHAT_WINDOWS do
		windowList = {GetChatWindowChannels(tab)};
		for i = 1, #windowList, 2 do
			if (windowList[i+1] == 0) then
				local lname = string.lower(windowList[i]);
				local sdata = _channelSavedList[lname];
				if (sdata and sdata.password) then
					if (not sdata.tabs) then
						sdata.tabs = {};
					end
					table.insert(sdata.tabs, tab);
				end
			end
		end
	end
	-- Save list
	ChatmanagerState.Servers[_serverName].Characters[_playerName].Channels = _channelSavedList;
	-- Record chat logging state if needed
	if (ChatmanagerState.Options.ChatLog == 1) then
		if (LoggingChat()) then
			ChatmanagerState.Options.ChatLog = 2;
		end
	end

	-- Save last sticky type
	ChatmanagerState.Servers[_serverName].Characters[_playerName].LastStickyType = ChatFrameEditBox:GetAttribute("stickyType");
end


------------------------------------------------------------------------------
-- Utility functions
------------------------------------------------------------------------------
function Chatmanager.GetNextParam(text)
	-- Extracts the next parameter out of the passed text, and returns it and
	-- the rest of the string
	for param, remain in string.gmatch(text, "([^%s]+)%s+(.*)") do
		return param, remain;
	end
	return text;
end


function Chatmanager.RebuildPlayerFilteredList(cdata)
	-- Rebuilds the filteredList and playerCount properties for the indicated
	-- channel data entry
	if (cdata and cdata.playerList) then
		local list = cdata.filteredList;
		if (list) then
			for index in pairs(list) do
				list[index] = nil;
			end
		else
			list = {};
		end
		for name in pairs(cdata.playerList) do
			table.insert(list, name);
		end
		table.sort(list);
		cdata.filteredList = list;
	end
end


function Chatmanager.ResetOrdering()
	-- Resets all channel display indices to their actual client indices, and
	-- removes all saved indices so that new channels will be joined at their
	-- client index by default.
	_channelIndexLookupByDisplayIndex = {};
	for index, cdata in ipairs(_channelList) do
		if (cdata.name) then
			cdata.displayIndex = index;
			_channelIndexLookupByDisplayIndex[index] = index;
		end
	end
	for sname, sdata in pairs(_channelSavedList) do
		sdata.index = nil;
	end
	Chatmanager.ChannelList_Update();
end


function Chatmanager.ResetColors()
	CloseDropDownMenus();
	ColorPickerFrame:Hide();
	for index, cdata in ipairs(_channelList) do
		if (cdata.name and cdata.suspended) then
			local lname = string.lower(cdata.name);
			local sdata = _channelSavedList[lname];
			if (sdata and sdata.color) then
				sdata.color.r = _defaultChannelColor.r;
				sdata.color.g = _defaultChannelColor.g;
				sdata.color.b = _defaultChannelColor.b;
			end
		end
		ChangeChatColor("CHANNEL"..index, _defaultChannelColor.r, _defaultChannelColor.g, _defaultChannelColor.b);
	end
end


function Chatmanager.ResetWindows()
	-- Menu button
	ChatmanagerMenuButtonFrame:ClearAllPoints();
	if (_hideChatButtons == 1) then
		ChatmanagerMenuButtonFrame:SetPoint("TOPRIGHT", "ChatFrame1", "TOPRIGHT");
	else
		ChatmanagerMenuButtonFrame:SetPoint("BOTTOM", "ChatFrameMenuButton", "TOP");
	end
	ChatmanagerMenuButtonFrame:SetUserPlaced(false);
	-- Chat menu (if modified)
	if (_buttonsHidden == 1) then
		ChatMenu:ClearAllPoints();
		if (_hideChatButtons == 1) then
			ChatMenu:SetPoint("BOTTOMLEFT", "ChatmanagerMenuButtonFrame", "TOPRIGHT");
		else
			ChatMenu:SetPoint("BOTTOMLEFT", "ChatFrameMenuButton", "TOPRIGHT");
		end
		ChatMenu:SetUserPlaced(false);
	end
	-- Chatmanager frame
	ChatmanagerFrame:ClearAllPoints();
	ChatmanagerFrame:SetPoint("TOPLEFT", nil, "TOPLEFT", 0, -104);
	-- Chatmanager options frame
	ChatmanagerOptionsFrame:ClearAllPoints();
	ChatmanagerOptionsFrame:SetPoint("TOPLEFT", ChatmanagerFrame, "TOPRIGHT", -4, 0);
end


function Chatmanager.MoveChannelTo(index, displayIndex)
	-- Moves the selected channel to the indicated display index
	if (index and displayIndex) then
		local cdata = _channelList[index];
		if (cdata and cdata.name and cdata.displayIndex) then
			local replaceIndex = _channelIndexLookupByDisplayIndex[displayIndex];
			if (replaceIndex) then
				local pdata = _channelList[replaceIndex];
				if (pdata and pdata.name and pdata.displayIndex) then
					pdata.displayIndex = cdata.displayIndex;
				end
			end
			_channelIndexLookupByDisplayIndex[displayIndex] = index;
			_channelIndexLookupByDisplayIndex[cdata.displayIndex] = replaceIndex;
			cdata.displayIndex = displayIndex;
		end
	end
end


function Chatmanager.GetPlayerOrGuildNote(playerName)
	-- Returns either the note assigned by the user to this playerName, or if
	-- none is assigned, and the user is in a guild, then it returns the
	-- player's guild public note (if there is one or "" otherwise) or nil if
	-- the player is not a guild member.
	local note = ChatmanagerState.Servers[_serverName].Notes[playerName];
	if (not note and IsInGuild() and (ChatmanagerState.Options.UseGuildNote == 1)) then
		guildies = GetNumGuildMembers(true);
		for i = 1, guildies do
			local name, _, _, _, _, _, publicnote = GetGuildRosterInfo(i);
			if (name == playerName) then
				if (publicnote) then
					note = publicnote;
				else
					note = "";
				end
				break;
			end
		end
	end
	return note;
end


------------------------------------------------------------------------------
-- ChatmanagerFrame UI functions
------------------------------------------------------------------------------
function Chatmanager.Frame_OnLoad()
	-- Make our frame closable with the ESC key
	tinsert(UISpecialFrames, "ChatmanagerFrame");

	-- Set title and version text
	ChatmanagerFrameTitleText:SetText(CHATMANAGER_TEXT.FRAME_TITLE);
	ChatmanagerFrameVersionText:SetText("v"..CHATMANAGER_VERSION.." by Cirk");
	ChatmanagerFrameVersionText:SetAlpha(0.5);

	-- Set text for buttons, tooltips, etc.
	ChatmanagerChannelSettingsFramePasswordText:SetText(CHATMANAGER_TEXT.PASSWORD_TITLE);
	ChatmanagerChannelSettingsFrameForcePasswordText:SetText(CHATMANAGER_TEXT.ALWAYS_USE_PASSWORD_TITLE);
	ChatmanagerChannelSettingsFrameForcePasswordCheckButton.tooltipText = CHATMANAGER_TEXT.ALWAYS_USE_PASSWORD_TOOLTIP;
	ChatmanagerChannelSettingsFrameStickyText:SetText(CHATMANAGER_TEXT.CHANNEL_STICKY_TITLE);
	ChatmanagerChannelSettingsFrameStickyCheckButton.tooltipText = CHATMANAGER_TEXT.CHANNEL_STICKY_TOOLTIP;
	ChatmanagerChannelSettingsFrameColorText:SetText(CHATMANAGER_TEXT.CHANNEL_COLOR_TITLE);
	ChatmanagerChannelSettingsFrameColorSwatch.tooltipText = CHATMANAGER_TEXT.CHANNEL_COLOR_TOOLTIP;
	ChatmanagerChannelSettingsFramePasswordSetButtonText:SetText(CHATMANAGER_TEXT.PASSWORD_SET_BUTTON);
	ChatmanagerFrameOptionsButtonText:SetText(CHATMANAGER_TEXT.OPTIONS_BUTTON);
	ChatmanagerPlayerListFrameLoadingText:SetText(CHATMANAGER_TEXT.LOADING_TEXT);
end


function Chatmanager.Frame_OnShow()
	Chatmanager.UpdateChannelList(1, 1);
	Chatmanager.ChannelList_Update();
	Chatmanager.UpdateChannelListSelection();
	Chatmanager.Frame_UpdatePlayerList();
end


function Chatmanager.Frame_OnHide()
	CloseDropDownMenus();
	Chatmanager.ChannelList_Update();
end


function Chatmanager.Frame_UpdateChannelList(buttonID)
	-- Called when the channel list has changed (buttonID is nil) or when a
	-- new channel has been selected (buttonID set to the display index of the
	-- channel selected).
	if (_debugFrame) then
		if (buttonID) then
			_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Frame_UpdateChannelList("..buttonID..")");
		else
			_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Frame_UpdateChannelList()");
		end
	end
	if (ChatmanagerFrame:IsShown()) then
		if (buttonID) then
			CloseDropDownMenus();
			_selectedChannelDisplayIndex = tonumber(buttonID);
		end
	elseif (buttonID) then
		_selectedChannelDisplayIndex = tonumber(buttonID);
		return;
	end

	-- Update the channel listing
	local lastSelectedChannelIndex = _selectedChannelIndex;
	_selectedChannelIndex = nil;
	for i = 1, CHATMANAGER_LAST_CHANNEL do
		local button = getglobal("ChatmanagerChannelListFrameButton"..i);
		local nameText = getglobal("ChatmanagerChannelListFrameButton"..i.."NameText");
		local numberText = getglobal("ChatmanagerChannelListFrameButton"..i.."NumberText");
		local highlight = getglobal("ChatmanagerChannelListFrameButton"..i.."Highlight");
		local index = _channelIndexLookupByDisplayIndex[i];
		local cdata = _channelList[index or 0];
		if (cdata and cdata.name) then
			local count = cdata.playerCount or 0;
			if ((count == 0) and cdata.filteredList) then
				count = #cdata.filteredList;
			end
			local name = cdata.fullName or cdata.name;
			if ((count > 0) and (cdata.type == 0)) then
				nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_FORMAT, name, count);
			elseif (not cdata.suspended) then
				nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_NOPLAYERS_FORMAT, name);
			else
				nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_SUSPENDED_FORMAT, name);
			end
			if (i == _selectedChannelDisplayIndex) then
				_selectedChannelIndex = index;
				button:LockHighlight();
			else
				button:UnlockHighlight();
			end
			local alpha = 1;
			if (_channelList[index].suspended) then
				alpha = _suspendedAlpha;
			end
			local info = ChatTypeInfo["CHANNEL"..index];
			nameText:SetTextColor(info.r, info.g, info.b, alpha);
			numberText:SetTextColor(info.r, info.g, info.b, alpha);
			highlight:SetVertexColor(info.r, info.g, info.b, _highlightAlpha*alpha);
			Chatmanager.Frame_UpdateChannelVoice(i, cdata.voiceEnabled, cdata.voiceActive);
		else
			nameText:SetText("");
			if (i == _selectedChannelDisplayIndex) then
				button:LockHighlight();
			else
				button:UnlockHighlight();
			end
			nameText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
			numberText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
			highlight:SetVertexColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b, _highlightAlpha);
			Chatmanager.Frame_UpdateChannelVoice(i, nil, nil);
		end
	end

	-- Update the channel settings and if the selected channel has changed (or
	-- has been clicked on again) then refresh the player list
	if (_selectedChannelIndex) then
		local title = string.format(CHATMANAGER_TEXT.FRAME_SETTINGS_TITLE_FORMAT, _channelList[_selectedChannelIndex].name);
		local info = ChatTypeInfo["CHANNEL".._selectedChannelIndex];
		local password = _channelList[_selectedChannelIndex].password;
		local sticky = _channelList[_selectedChannelIndex].isSticky;
		local forcePassword = _channelList[_selectedChannelIndex].forcePassword;
		ChatmanagerChannelSettingsFrameTitle:SetText(title);
		ChatmanagerChannelSettingsFrameColorSwatchNormalTexture:SetVertexColor(info.r, info.g, info.b);
		ChatmanagerChannelSettingsFrameColorSwatch.channelID = _selectedChannelIndex;
		ChatmanagerChannelSettingsFrameColorSwatch:Enable();
		if (_channelList[_selectedChannelIndex].type == 0) then
			ChatmanagerChannelSettingsFramePasswordText:SetTextColor(_colorTextEnabled.r, _colorTextEnabled.g, _colorTextEnabled.b);
			ChatmanagerChannelSettingsFrameForcePasswordText:SetTextColor(_colorTextEnabled.r, _colorTextEnabled.g, _colorTextEnabled.b);
			ChatmanagerChannelSettingsFrameEditBox:Show();
			if (buttonID or (lastSelectedChannelIndex ~= _selectedChannelIndex)) then
				local text = ChatmanagerChannelSettingsFrameEditBox:GetText();
				if (password) then
					if (password ~= text) then
						ChatmanagerChannelSettingsFrameEditBox:SetText(password);
						ChatmanagerChannelSettingsFrameEditBox.textSet = 1;
					end
				elseif (text ~= "") then
					ChatmanagerChannelSettingsFrameEditBox:SetText("");
					ChatmanagerChannelSettingsFrameEditBox.textSet = 1;
				end
				ChatmanagerChannelSettingsFrameEditBox:ClearFocus();
				ChatmanagerChannelSettingsFramePasswordSetButton:Disable();
			end
			ChatmanagerChannelSettingsFrameForcePasswordCheckButton:Enable();
			if (forcePassword == 1) then
				ChatmanagerChannelSettingsFrameForcePasswordCheckButton:SetChecked(1);
			else
				ChatmanagerChannelSettingsFrameForcePasswordCheckButton:SetChecked(0);
			end
		else
			ChatmanagerChannelSettingsFramePasswordText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
			ChatmanagerChannelSettingsFrameForcePasswordText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
			ChatmanagerChannelSettingsFrameEditBox:Hide();
			ChatmanagerChannelSettingsFramePasswordSetButton:Disable();
			ChatmanagerChannelSettingsFrameForcePasswordCheckButton:Disable();
			ChatmanagerChannelSettingsFrameForcePasswordCheckButton:SetChecked(0);
		end
		ChatmanagerChannelSettingsFrameStickyCheckButton:Enable();
		if (sticky == 1) then
			ChatmanagerChannelSettingsFrameStickyCheckButton:SetChecked(1);
		else
			ChatmanagerChannelSettingsFrameStickyCheckButton:SetChecked(0);
		end
		ChatmanagerChannelSettingsFrameColorText:SetTextColor(_colorTextEnabled.r, _colorTextEnabled.g, _colorTextEnabled.b);
		ChatmanagerChannelSettingsFrameStickyText:SetTextColor(_colorTextEnabled.r, _colorTextEnabled.g, _colorTextEnabled.b);
		-- Refresh the player list (if not already in progress)
		if (buttonID or (lastSelectedChannelIndex ~= _selectedChannelIndex)) then
			if (not _channelList[_selectedChannelIndex].suspended) then
				Chatmanager.RequestPlayerList(_selectedChannelIndex, false, true);
			end
		end
	else
		ChatmanagerChannelSettingsFrameTitle:SetText(CHATMANAGER_TEXT.SETTINGS_TITLE_UNSELECTED);
		ChatmanagerChannelSettingsFrameColorSwatchNormalTexture:SetVertexColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
		ChatmanagerChannelSettingsFrameColorSwatch.channelID = nil;
		ChatmanagerChannelSettingsFrameColorSwatch:Disable();
		ChatmanagerChannelSettingsFrameEditBox:Hide();
		ChatmanagerChannelSettingsFramePasswordSetButton:Disable();
		ChatmanagerChannelSettingsFrameStickyCheckButton:Disable();
		ChatmanagerChannelSettingsFrameStickyCheckButton:SetChecked(0);
		ChatmanagerChannelSettingsFrameForcePasswordCheckButton:Disable();
		ChatmanagerChannelSettingsFrameForcePasswordCheckButton:SetChecked(0);
		ChatmanagerChannelSettingsFramePasswordText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
		ChatmanagerChannelSettingsFrameForcePasswordText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
		ChatmanagerChannelSettingsFrameColorText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
		ChatmanagerChannelSettingsFrameStickyText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
	end
end


function Chatmanager.Frame_UpdateChannelPlayerCounts(cdata)
	-- Updates just the player counts shown next to the channel names.  If the
	-- optional cdata is provided then only that channel will be updated.
	if (cdata and cdata.name and cdata.displayIndex) then
		local nameText = getglobal("ChatmanagerChannelListFrameButton"..cdata.displayIndex.."NameText");
		if (nameText) then
			local count = cdata.playerCount or 0;
			if ((count == 0) and cdata.filteredList) then
				count = #cdata.filteredList;
			end
			local name = cdata.fullName or cdata.name;
			if ((count > 0) and (cdata.type == 0)) then
				nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_FORMAT, name, count);
			elseif (not cdata.suspended) then
				nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_NOPLAYERS_FORMAT, name);
			else
				nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_SUSPENDED_FORMAT, name);
			end
		end
	else
		for i = 1, CHATMANAGER_LAST_CHANNEL do
			local nameText = getglobal("ChatmanagerChannelListFrameButton"..i.."NameText");
			local index = _channelIndexLookupByDisplayIndex[i];
			local cdata = _channelList[index or 0];
			if (cdata and cdata.name) then
				local count = cdata.playerCount or 0;
				if ((count == 0) and cdata.filteredList) then
					count = #cdata.filteredList;
				end
				local name = cdata.fullName or cdata.name;
				if ((count > 0) and (cdata.type == 0)) then
					nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_FORMAT, name, count);
				elseif (not cdata.suspended) then
					nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_NOPLAYERS_FORMAT, name);
				else
					nameText:SetFormattedText(CHATMANAGER_TEXT.FRAME_CHANNEL_NAME_SUSPENDED_FORMAT, name);
				end
			end
		end
	end
end


function Chatmanager.Frame_UpdatePlayerList(self)
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Frame_UpdatePlayerList()");
	end
	if (not ChatmanagerFrame:IsShown()) then
		return;
	end
	local offset = FauxScrollFrame_GetOffset(ChatmanagerPlayerListScrollFrame);
	local currentOffset = offset;
	local playerList = nil;
	local filteredList = nil;
	local playerCount = 0;
	local showCount = 0;
	local active = nil;

	if (_selectedChannelIndex) then
		local cdata = _channelList[_selectedChannelIndex];
		if (cdata and cdata.playerList and cdata.filteredList and cdata.playerCount) then
			playerList = cdata.playerList;
			filteredList = cdata.filteredList;
			playerCount = #filteredList;
			showCount = cdata.playerCount;
			active = not cdata.suspended;
		end
	end

	if ((offset + CHATMANAGER_PLAYER_LIST_BUTTON_COUNT) > playerCount) then
		offset = playerCount - CHATMANAGER_PLAYER_LIST_BUTTON_COUNT;
	end
	if (offset < 0) then
		offset = 0;
	end
	if (offset ~= currentOffset) then
		FauxScrollFrame_SetOffset(ChatmanagerPlayerListScrollFrame, offset);
		ChatmanagerPlayerListScrollFrameScrollBar:SetValue(offset * CHATMANAGER_PLAYER_LIST_BUTTON_HEIGHT);
	end

	-- Update title
	if (showCount > 0) then
		ChatmanagerPlayerListFrameTitle:SetFormattedText(CHATMANAGER_TEXT.FRAME_PLAYERS_TITLE_FORMAT, showCount);
	else
		ChatmanagerPlayerListFrameTitle:SetText(CHATMANAGER_TEXT.FRAME_PLAYERS_TITLE_NONE);
	end

	-- Update buttons
	for i = 1, CHATMANAGER_PLAYER_LIST_BUTTON_COUNT do
		local button = getglobal("ChatmanagerPlayerListFrameButton"..i);
		local index = i + offset;
		if (index <= playerCount) then
			local rankIcon = getglobal("ChatmanagerPlayerListFrameButton"..i.."RankTexture");
			local nameText = getglobal("ChatmanagerPlayerListFrameButton"..i.."NameText");
			local nameHighlight = getglobal("ChatmanagerPlayerListFrameButton"..i.."Highlight");
			local actionHighlight = getglobal("ChatmanagerPlayerListFrameButton"..i.."ActionHighlight");
			local pname = filteredList[index];
			local pdata = playerList[pname];
			button.pindex = index;
			button.pname = pname;
			if (pdata.isOwner) then
				rankIcon:SetTexture("Interface\\GroupFrame\\UI-Group-LeaderIcon");
				rankIcon:Show();
				nameText:SetFormattedText(CHATMANAGER_TEXT.PLAYER_NAME_NORMAL_FORMAT, pname);
				nameText:SetTextColor(_playerOwner.r, _playerOwner.g, _playerOwner.b);
				nameHighlight:SetVertexColor(_playerOwner.r, _playerOwner.g, _playerOwner.b, _highlightAlpha);
				actionHighlight:SetVertexColor(_playerOwner.r, _playerOwner.g, _playerOwner.b, _highlightAlpha);
				button.tooltipText = string.format(CHATMANAGER_TEXT.PLAYER_OWNER_TOOLTIP_FORMAT, pname);
			elseif (pdata.isModerator) then
				rankIcon:SetTexture("Interface\\GroupFrame\\UI-Group-AssistantIcon");
				rankIcon:Show();
				nameText:SetFormattedText(CHATMANAGER_TEXT.PLAYER_NAME_NORMAL_FORMAT, pname);
				nameText:SetTextColor(_playerModerator.r, _playerModerator.g, _playerModerator.b);
				nameHighlight:SetVertexColor(_playerModerator.r, _playerModerator.g, _playerModerator.b, _highlightAlpha);
				actionHighlight:SetVertexColor(_playerModerator.r, _playerModerator.g, _playerModerator.b, _highlightAlpha);
				button.tooltipText = string.format(CHATMANAGER_TEXT.PLAYER_MODERATOR_TOOLTIP_FORMAT, pname);
			elseif (pdata.isMuted) then
				rankIcon:Hide();
				nameText:SetFormattedText(CHATMANAGER_TEXT.PLAYER_NAME_MUTED_FORMAT, pname);
				nameText:SetTextColor(_playerMuted.r, _playerMuted.g, _playerMuted.b);
				nameHighlight:SetVertexColor(_playerMuted.r, _playerMuted.g, _playerMuted.b, _highlightAlpha);
				button.tooltipText = string.format(CHATMANAGER_TEXT.PLAYER_MUTED_TOOLTIP_FORMAT, pname);
			else
				rankIcon:Hide();
				nameText:SetFormattedText(CHATMANAGER_TEXT.PLAYER_NAME_NORMAL_FORMAT, pname);
				nameText:SetTextColor(_playerNormal.r, _playerNormal.g, _playerNormal.b);
				nameHighlight:SetVertexColor(_playerNormal.r, _playerNormal.g, _playerNormal.b, _highlightAlpha);
				actionHighlight:SetVertexColor(_playerNormal.r, _playerNormal.g, _playerNormal.b, _highlightAlpha);
				button.tooltipText = nil;
			end
			local pnote = Chatmanager.GetPlayerOrGuildNote(pname);
			if (pnote and (pnote ~= "")) then
				button.tooltipText2 = pnote;
			else
				button.tooltipText2 = nil;
			end
			if (_selectedPlayer == pname) then
				if (not button.noHighlight) then
					button:LockHighlight();
				end
				button.selected = 1;
			else
				button:UnlockHighlight();
				button.selected = nil;
			end
			Chatmanager.Frame_UpdatePlayerVoice(i, pdata.voiceEnabled, pdata.voiceActive, pdata.voiceMuted);
			button:Show();
		else
			button.pindex = nil;
			button.pname = nil;
			button.selected = nil;
			button:Hide();
		end
	end

	if ((playerCount == 0) and active) then
		ChatmanagerPlayerListFrameLoading:Show();
	else
		ChatmanagerPlayerListFrameLoading:Hide();
	end

	-- Update scrollbar
	FauxScrollFrame_Update(ChatmanagerPlayerListScrollFrame, playerCount, CHATMANAGER_PLAYER_LIST_BUTTON_COUNT, CHATMANAGER_PLAYER_LIST_BUTTON_HEIGHT,
							"ChatmanagerPlayerListFrameButton", 142, 158);
end


function Chatmanager.Frame_StickyOnClick(self)
	if (_selectedChannelIndex) then
		local cdata = _channelList[_selectedChannelIndex];
		if (cdata and cdata.name) then
			if (self:GetChecked()) then
				cdata.isSticky = 1;
			else
				cdata.isSticky = 0;
			end
		end
	end
end


function Chatmanager.Frame_ForcePasswordOnClick(self)
	if (_selectedChannelIndex) then
		local cdata = _channelList[_selectedChannelIndex];
		if (cdata and cdata.name) then
			if (self:GetChecked()) then
				cdata.forcePassword = 1;
			else
				cdata.forcePassword = 0;
			end
		end
	end
end


function Chatmanager.Frame_SetChatColor(previous)
	local r, g, b = ColorPickerFrame:GetColorRGB();
	if (previous) then
		r = previous.r;
		g = previous.g;
		b = previous.b;
	end
	ChangeChatColor(UIDROPDOWNMENU_MENU_VALUE, r, g, b);
	if (_selectedChannelIndex) then
		local cdata = _channelList[_selectedChannelIndex];
		if (cdata and cdata.name) then
			if (cdata.suspended) then
				local lname = string.lower(cdata.name);
				local sdata = _channelSavedList[lname];
				if (sdata and sdata.color) then
					sdata.color.r = r;
					sdata.color.g = g;
					sdata.color.b = b;
				end
			end
			if (ChatmanagerFrame:IsShown() and _selectedChannelDisplayIndex) then
				ChatmanagerChannelSettingsFrameColorSwatchNormalTexture:SetVertexColor(r, g, b);
				local alpha = 1;
				if (cdata.suspended) then
					alpha = _suspendedAlpha;
				end
				local nameText = getglobal("ChatmanagerChannelListFrameButton".._selectedChannelDisplayIndex.."NameText");
				local numberText = getglobal("ChatmanagerChannelListFrameButton".._selectedChannelDisplayIndex.."NumberText");
				local highlight = getglobal("ChatmanagerChannelListFrameButton".._selectedChannelDisplayIndex.."Highlight");
				nameText:SetTextColor(r, g, b, alpha);
				numberText:SetTextColor(r, g, b, alpha);
				highlight:SetVertexColor(r, g, b, _highlightAlpha*alpha);
			end
		end
	end
end


function Chatmanager.Frame_OnUpdate()
	-- Checks for any pending names in the various player lists, and if any
	-- are found they are removed.  Also resets the userRequest flag for
	-- each channel if the request has expired.
	local time = GetTime();
	if (_playerListUpdateTimeout and (time > _playerListUpdateTimeout)) then
		local nextUpdate = nil;
		for index, cdata in ipairs(_channelList) do
			if (cdata.pending) then
				if (time > cdata.pending) then
					local playerList = cdata.playerList;
					if (playerList) then
						for name, pdata in pairs(playerList) do
							if (pdata and pdata.pending) then
								playerList[name] = nil;
							end
						end
					end
					cdata.pending = nil;
					if (cdata.request) then
						cdata.request = time + CHATMANAGER_PLAYER_REQUEST_DELAY;
					end
					Chatmanager.RebuildPlayerFilteredList(cdata);
					Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
					if (index == _selectedChannelIndex) then
						Chatmanager.Frame_UpdatePlayerList();
					end
					cdata.userRequest = nil;
				end
			elseif (cdata.request) then
				if (time > cdata.request) then
					cdata.request = nil;
					Chatmanager.RequestPlayerList(index, cdata.nextUserRequest, (index == _selectedChannelIndex) or (cdata.type ~= 0));
					cdata.nextUserRequest = nil;
				end
			end
			if (cdata.pending) then
				if (not nextUpdate or (nextUpdate > cdata.pending)) then
					nextUpdate = cdata.pending;
				end
			elseif (cdata.request) then
				if (not nextUpdate or (nextUpdate > cdata.request)) then
					nextUpdate = cdata.request;
				end
			end
		end
		_playerListUpdateTimeout = nextUpdate;
	end
	if (_channelListUpdateTimeout and (time > _channelListUpdateTimeout)) then
		_channelListUpdateTimeout = nil;
		Chatmanager.UpdateChannelList(1, 1);
	end
	if (_channelConnects) then
		for name, sdata in pairs(_channelConnects) do
			if (sdata.password and (sdata.password ~= "")) then
				sdata.newPassword = sdata.password;
				sdata.password = nil;
				if (_debugFrame) then
					_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Attempting to reconnect to "..sdata.name.." with password "..sdata.newPassword);
				end
				_original_JoinPermanentChannel(sdata.name, sdata.newPassword);
				_channelListUpdateTimeout = GetTime() + CHATMANAGER_CHANNEL_JOIN_UPDATE_DELAY;
			end
		end
		_channelConnects = nil;
	end
	if (not _playerListUpdateTimeout and not _channelListUpdateTimeout) then
		_thisFrame:SetScript("OnUpdate", nil);
	end
end


function Chatmanager.Frame_ChannelButtonOnClick(self)
	-- Called when the user left or right clicks on a channel in the channel
	-- list.
	-- Close any open menus and close the color picker if it is open too.
	CloseDropDownMenus();
	if (ColorPickerFrame:IsShown()) then
		if (ColorPickerFrame.cancelFunc) then
			ColorPickerFrame.cancelFunc(ColorPickerFrame.previousValues);
		end
		ColorPickerFrame:Hide();
	end
	-- Set frame to show this channel (or refresh channel if it is already on
	-- this one).
	Chatmanager.Frame_UpdateChannelList(self:GetID());
	local cdata;
	if (_selectedChannelIndex) then
		cdata = _channelList[_selectedChannelIndex];
		if (cdata and cdata.name and cdata.channelInfoID) then
			if (not cdata.suspended) then
				ChannelList_UpdateHighlight(cdata.channelInfoID);
				ChannelFrame.updating = cdata.channelInfoID;
				SetSelectedDisplayChannel(cdata.channelInfoID);
				ChannelRoster_Update(cdata.channelInfoID);
			else
				ChannelList_UpdateHighlight(0);
				SetSelectedDisplayChannel(cdata.channelInfoID);
				ChannelRoster_Update(0);
			end
		end
	end
	Chatmanager.Frame_UpdatePlayerList();

	-- Process menu display
	if (arg1 == "RightButton") then
		if (cdata and cdata.name) then
			ChatmanagerDropDown.thisMode = "channel";
			ChatmanagerDropDown.thisName = cdata.name;
			ChatmanagerDropDown.thisButton = self;
			ChatmanagerDropDown.initialize = Chatmanager.Frame_DropdownTargetInitialize;
			ChatmanagerDropDown.displayMode = "MENU";
			ChatmanagerDropDown.point = "TOPLEFT";
			ChatmanagerDropDown.relativePoint = "BOTTOMLEFT";
			ToggleDropDownMenu(1, nil, ChatmanagerDropDown, "cursor");
		else
			ChatmanagerDropDown.thisMode = "newchannel";
			ChatmanagerDropDown.thisName = nil;
			ChatmanagerDropDown.thisButton = self;
			ChatmanagerDropDown.initialize = Chatmanager.Frame_DropdownTargetInitialize;
			ChatmanagerDropDown.displayMode = "MENU";
			ChatmanagerDropDown.point = "TOPLEFT";
			ChatmanagerDropDown.relativePoint = "BOTTOMLEFT";
			ToggleDropDownMenu(1, nil, ChatmanagerDropDown, "cursor");
		end
	end
end


function Chatmanager.Frame_PlayerButtonPreClick(self, action)
	-- Called for the <PreClick> script handler when the user left or right
	-- clicks on a player in the player list.
	local button = self;
	if (action) then
		-- If called from the action button, then use its parent for
		-- information, rather than the action button itself
		button = self:GetParent();
	end
	if (not button) then
		return;
	end

	-- Close any open menus and close the color picker if it is open too.
	CloseDropDownMenus();
	if (ColorPickerFrame:IsShown()) then
		if (ColorPickerFrame.cancelFunc) then
			ColorPickerFrame.cancelFunc(ColorPickerFrame.previousValues);
		end
		ColorPickerFrame:Hide();
	end

	-- In WoW 2.0, addons cannot call the TargetUnit function directly to
	-- target a specific unit or player, so here we use the SecureActionButton
	-- template's attribute functions with a macro to make use of the
	-- /targetexact (2.3) or /target (2.2) slash command.
	if ((arg1 == "LeftButton") and action) then
		if (button.pname) then
			if (not InCombatLockdown()) then
				self:SetAttribute("type1", "macro");
				if (_hasTargetExact) then
					self:SetAttribute("macrotext", "/targetexact [nocombat] "..button.pname);
				else
					self:SetAttribute("macrotext", "/target [nocombat] "..button.pname);
				end
			end
		end
	elseif (arg1 == "RightButton") then
		ChatmanagerDropDown.thisMode = "player";
		ChatmanagerDropDown.thisName = button.pname;
		ChatmanagerDropDown.thisButton = button;
		ChatmanagerDropDown.initialize = Chatmanager.Frame_DropdownTargetInitialize;
		ChatmanagerDropDown.displayMode = "MENU";
		ChatmanagerDropDown.point = "TOPLEFT";
		ChatmanagerDropDown.relativePoint = "BOTTOMLEFT";
		ToggleDropDownMenu(1, nil, ChatmanagerDropDown, "cursor");
	end
end


function Chatmanager.Frame_PlayerButtonPostClick(self)
	-- Remove attributes from the buttons when no longer needed
	self:SetAttribute("type1", nil);
	self:SetAttribute("macrotext", nil);
end


function Chatmanager.Frame_DropdownTargetInitialize()
	local info;
	if (ChatmanagerDropDown.thisMode == "channel") then
		if (UIDROPDOWNMENU_MENU_LEVEL == 1) then
			local cdata = _channelList[_selectedChannelIndex or 0];
			if (IsVoiceChatEnabled() and cdata.name and (cdata.type == 0)) then
				if (not cdata.voiceEnabled and cdata.amOwner and cdata.channelInfoID) then
					-- Need to be an owner to activate voice on a channel
					info = UIDropDownMenu_CreateInfo();
					info.text = CHATMANAGER_TEXT.ENABLE_VOICE_OPTION;
					info.value = CHATMANAGER_TEXT.ENABLE_VOICE_OPTION;
					info.notCheckable = 1;
					info.func = Chatmanager.Frame_DropdownSelect;
					UIDropDownMenu_AddButton(info);
				elseif (cdata.voiceEnabled and not cdata.voiceActive and cdata.channelInfoID) then
					-- Enable voice chat on this channel
					info = UIDropDownMenu_CreateInfo();
					info.text = CHATMANAGER_TEXT.VOICE_CHAT_OPTION;
					info.value = CHATMANAGER_TEXT.VOICE_CHAT_OPTION;
					info.notCheckable = 1;
					info.func = Chatmanager.Frame_DropdownSelect;
					if (cdata.voiceActive) then
						info.disabled = 1;
					end
					UIDropDownMenu_AddButton(info);
				end
			end
			-- Move channel to
			info = UIDropDownMenu_CreateInfo();
			info.text = CHATMANAGER_TEXT.MOVE_TO_OPTION;
			info.notCheckable = 1;
			info.hasArrow = 1;
			UIDropDownMenu_AddButton(info);
			-- Text color
			info = UIDropDownMenu_CreateInfo();
			info.text = CHATMANAGER_TEXT.TEXT_COLOR_OPTION;
			info.value = CHATMANAGER_TEXT.TEXT_COLOR_OPTION;
			if (_selectedChannelIndex) then
				info.textR = ChatTypeInfo["CHANNEL".._selectedChannelIndex].r;
				info.textG = ChatTypeInfo["CHANNEL".._selectedChannelIndex].g;
				info.textB = ChatTypeInfo["CHANNEL".._selectedChannelIndex].b;
			else
				info.disabled = 1;
			end
			info.notCheckable = 1;
			info.func = Chatmanager.Frame_DropdownSelect;
			UIDropDownMenu_AddButton(info);
			-- Leave channel
			info = UIDropDownMenu_CreateInfo();
			info.text = CHATMANAGER_TEXT.LEAVE_OPTION;
			info.value = CHATMANAGER_TEXT.LEAVE_OPTION;
			if (_selectedChannelIndex) then
				-- Note that it doesn't appear to be possible to leave suspended
				-- channels, so treat them like bad data.
				if (not cdata or not cdata.name or cdata.suspended) then
					info.disabled = 1;
				end
			end
			info.notCheckable = 1;
			info.func = Chatmanager.Frame_DropdownSelect;
			UIDropDownMenu_AddButton(info);
			-- Cancel menu
			info = UIDropDownMenu_CreateInfo();
			info.text = CANCEL;
			info.notCheckable = 1;
			info.func = Chatmanager.Frame_DropdownCancel;
			UIDropDownMenu_AddButton(info);
		elseif (UIDROPDOWNMENU_MENU_LEVEL == 2) then
			for i = 1, CHATMANAGER_LAST_CHANNEL do
				info = UIDropDownMenu_CreateInfo();
				info.text = string.format(CHATMANAGER_TEXT.MOVE_TO_CHANNEL_FORMAT, i);
				info.value = i;
				if (not ChatmanagerDropDown.thisButton or (ChatmanagerDropDown.thisButton:GetID() == i)) then
					info.disabled = 1;
				end
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
			end
		end
	elseif (ChatmanagerDropDown.thisMode == "newchannel") then
		-- New channel
		info = UIDropDownMenu_CreateInfo();
		info.text = CHATMANAGER_TEXT.NEW_CHANNEL_OPTION;
		info.value = "";
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownSelect;
		UIDropDownMenu_AddButton(info);
		-- Previous channels
		for sname, sdata in pairs(_channelSavedList) do
			if (sdata and not _channelIndexLookupByName[sname]) then
				info = UIDropDownMenu_CreateInfo();
				info.text = string.format(CHATMANAGER_TEXT.REJOIN_CHANNEL_OPTION, sdata.name);
				info.value = sname;
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
			end
		end
		-- Cancel menu
		info = UIDropDownMenu_CreateInfo();
		info.text = CANCEL;
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownCancel;
		UIDropDownMenu_AddButton(info);
	elseif (ChatmanagerDropDown.thisMode == "player") then
		if (ChatmanagerDropDown.thisName) then
			info = UIDropDownMenu_CreateInfo();
			info.text = ChatmanagerDropDown.thisName;
			info.isTitle = 1;
			info.notCheckable = 1;
			UIDropDownMenu_AddButton(info);
		end
		-- Player Note
		info = UIDropDownMenu_CreateInfo();
		info.text = CHATMANAGER_TEXT.EDIT_PLAYER_NOTE;
		info.value = CHATMANAGER_TEXT.EDIT_PLAYER_NOTE;
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownSelect;
		UIDropDownMenu_AddButton(info);
		-- Who
		info = UIDropDownMenu_CreateInfo();
		info.text = WHO;
		info.value = WHO;
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownSelect;
		UIDropDownMenu_AddButton(info);
		-- Whisper
		info = UIDropDownMenu_CreateInfo();
		info.text = WHISPER;
		info.value = WHISPER;
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownSelect;
		UIDropDownMenu_AddButton(info);
		-- Invite
		info = UIDropDownMenu_CreateInfo();
		info.text = PARTY_INVITE;
		info.value = PARTY_INVITE;
		if (ChatmanagerDropDown.thisName == _playerName) then
			info.disabled = 1;
		elseif (GetNumRaidMembers() > 0) then
			if (IsRaidLeader() or IsRaidOfficer()) then
				for i = 1, 40 do
					local unit = "raid"..i;
					if (UnitInRaid(unit) and (UnitName(unit) == ChatmanagerDropDown.thisName)) then
						info.disabled = 1;
						break;
					end
				end
			else
				info.disabled = 1;
			end
		elseif (GetNumPartyMembers() > 0) then
			if (IsPartyLeader()) then
				for i = 1, 4 do
					local unit = "party"..i;
					if (UnitInParty(unit) and (UnitName(unit) == ChatmanagerDropDown.thisName)) then
						info.disabled = 1;
						break;
					end
				end
			else
				info.disabled = 1;
			end
		end
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownSelect;
		UIDropDownMenu_AddButton(info);
		-- Variables for the options to follow
		local cdata = nil;
		local pdata = nil;
		local moderator = nil;
		local notServer = nil;
		local voiceChat = IsVoiceChatEnabled();
		local isSelf = (ChatmanagerDropDown.thisName == _playerName);
		if (_selectedChannelIndex) then
			cdata = _channelList[_selectedChannelIndex];
			if (cdata and cdata.name) then
				moderator = cdata.amOwner or cdata.amModerator;
				if (cdata.type == 0) then
					notServer = true;
				end
				pdata = cdata.playerList[ChatmanagerDropDown.thisName];
			end
		end
		if (voiceChat and not isSelf) then
			-- Voice mute/unmute
			info = UIDropDownMenu_CreateInfo();
			if (IsMuted(ChatmanagerDropDown.thisName)) then
				info.text = CHATMANAGER_TEXT.VOICE_UNMUTE_PLAYER_OPTION;
				info.value = CHATMANAGER_TEXT.VOICE_UNMUTE_PLAYER_OPTION;
			else
				info.text = CHATMANAGER_TEXT.VOICE_MUTE_PLAYER_OPTION;
				info.value = CHATMANAGER_TEXT.VOICE_MUTE_PLAYER_OPTION;
			end
			info.notCheckable = 1;
			info.func = Chatmanager.Frame_DropdownSelect;
			UIDropDownMenu_AddButton(info);
		end
		if (notServer) then
			-- Voice silence/unsilence
			if (voiceChat and not isSelf and moderator) then
				info = UIDropDownMenu_CreateInfo();
				if (IsSilenced(ChatmanagerDropDown.thisName)) then
					info.text = CHATMANAGER_TEXT.VOICE_UNSILENCE_PLAYER_OPTION;
					info.value = CHATMANAGER_TEXT.VOICE_UNSILENCE_PLAYER_OPTION;
				else
					info.text = CHATMANAGER_TEXT.VOICE_SILENCE_PLAYER_OPTION;
					info.value = CHATMANAGER_TEXT.VOICE_SILENCE_PLAYER_OPTION;
				end
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
			end
			if (moderator and pdata and not pdata.isOwner and not pdata.isModerator) then
				-- Mute/Unmute
				info = UIDropDownMenu_CreateInfo();
				if (pdata.isMuted) then
					info.text = CHATMANAGER_TEXT.UNMUTE_PLAYER_OPTION;
					info.value = CHATMANAGER_TEXT.UNMUTE_PLAYER_OPTION;
				else
					info.text = CHATMANAGER_TEXT.MUTE_PLAYER_OPTION;
					info.value = CHATMANAGER_TEXT.MUTE_PLAYER_OPTION;
				end
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
				-- Kick
				info = UIDropDownMenu_CreateInfo();
				info.text = CHATMANAGER_TEXT.KICK_PLAYER_OPTION;
				info.value = CHATMANAGER_TEXT.KICK_PLAYER_OPTION;
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
				-- Ban
				info = UIDropDownMenu_CreateInfo();
				info.text = CHATMANAGER_TEXT.BAN_PLAYER_OPTION;
				info.value = CHATMANAGER_TEXT.BAN_PLAYER_OPTION;
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
			end
			if (moderator and pdata and not pdata.isOwner) then
				-- Set/unset moderator
				info = UIDropDownMenu_CreateInfo();
				if (pdata.isModerator) then
					info.text = CHATMANAGER_TEXT.UNSET_MODERATOR_OPTION;
					info.value = CHATMANAGER_TEXT.UNSET_MODERATOR_OPTION;
				else
					info.text = CHATMANAGER_TEXT.SET_MODERATOR_OPTION;
					info.value = CHATMANAGER_TEXT.SET_MODERATOR_OPTION;
				end
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
			end
			if (cdata and cdata.amOwner and not isSelf) then
				-- Set owner
				info = UIDropDownMenu_CreateInfo();
				info.text = CHATMANAGER_TEXT.SET_OWNER_OPTION;
				info.value = CHATMANAGER_TEXT.SET_OWNER_OPTION;
				info.notCheckable = 1;
				info.func = Chatmanager.Frame_DropdownSelect;
				UIDropDownMenu_AddButton(info);
			end
		end
		-- Cancel menu
		info = UIDropDownMenu_CreateInfo();
		info.text = CANCEL;
		info.notCheckable = 1;
		info.func = Chatmanager.Frame_DropdownCancel;
		UIDropDownMenu_AddButton(info);
	end
end


function Chatmanager.Frame_DropdownSelect(self)
	CloseDropDownMenus();
	if (ChatmanagerDropDown.thisMode == "channel") then
		if (self.value == CHATMANAGER_TEXT.TEXT_COLOR_OPTION) then
			Chatmanager.Frame_ColorSwatchOnClick(self, _selectedChannelIndex);
		elseif (self.value == CHATMANAGER_TEXT.LEAVE_OPTION) then
			if (_selectedChannelIndex) then
				local cdata = _channelList[_selectedChannelIndex];
				if (cdata and cdata.name) then
					local dialogFrame = StaticPopup_Show("CHATMANAGER_LEAVE_CONFIRM", cdata.name);
					if (dialogFrame) then
						dialogFrame.data = tostring(_selectedChannelIndex);
					end
				end
			end
		elseif (self.value == CHATMANAGER_TEXT.ENABLE_VOICE_OPTION) then
			local cdata = _channelList[_selectedChannelIndex or 0];
			if (cdata and cdata.name and cdata.channelInfoID) then
				DisplayChannelVoiceOn(cdata.channelInfoID);
			end
		elseif (self.value == CHATMANAGER_TEXT.VOICE_CHAT_OPTION) then
			-- Have to find the display channel ID corresponding to our channel
			local cdata = _channelList[_selectedChannelIndex or 0];
			if (cdata and cdata.name and cdata.channelInfoID) then
				SetActiveVoiceChannel(cdata.channelInfoID);
			end
		else
			local displayIndex = tonumber(self.value or 0);
			if (displayIndex and (displayIndex >= 1) and (displayIndex <= CHATMANAGER_LAST_CHANNEL)) then
				Chatmanager.MoveChannelTo(_selectedChannelIndex, displayIndex);
			end
			Chatmanager.Frame_UpdateChannelList(displayIndex);
			Chatmanager.ChannelList_Update();
		end
	elseif (ChatmanagerDropDown.thisMode == "newchannel") then
		if (self.value == "") then
			StaticPopup_Show("CHATMANAGER_NEW_CHANNEL");
		elseif (self.value) then
			local sdata = _channelSavedList[self.value];
			if (sdata) then
				sdata.index = _selectedChannelDisplayIndex;
				sdata.newPassword = sdata.password;
				sdata.password = nil;
				local zoneID = _original_JoinPermanentChannel(sdata.name, sdata.newPassword, DEFAULT_CHAT_FRAME:GetID(), 1);
				if (zoneID) then
					ChatFrame_RemoveChannel(DEFAULT_CHAT_FRAME, sdata.name);
					ChatFrame_AddChannel(DEFAULT_CHAT_FRAME, sdata.name);
				end
			end
		end
	elseif (ChatmanagerDropDown.thisMode == "player") then
		local cdata = nil;
		local moderator = nil;
		if (_selectedChannelIndex) then
			cdata = _channelList[_selectedChannelIndex];
			if (cdata) then
				moderator = cdata.amOwner or cdata.amModerator;
			end
		end
		if (self.value == WHO) then
			if (ChatmanagerDropDown.thisName) then
				SendWho("n-"..ChatmanagerDropDown.thisName);
			end
		elseif (self.value == WHISPER) then
			if (ChatmanagerDropDown.thisName) then
				ChatFrame_SendTell(ChatmanagerDropDown.thisName);
			end
		elseif (self.value == PARTY_INVITE) then
			if (ChatmanagerDropDown.thisName) then
				InviteUnit(ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.EDIT_PLAYER_NOTE) then
			if (ChatmanagerDropDown.thisName) then
				local dialogFrame = StaticPopup_Show("CHATMANAGER_PLAYER_NOTE", ChatmanagerDropDown.thisName);
				if (dialogFrame) then
					dialogFrame.data = ChatmanagerDropDown.thisName;
				end
			end
		elseif (self.value == CHATMANAGER_TEXT.VOICE_MUTE_PLAYER_OPTION) then
			if (ChatmanagerDropDown.thisName) then
				AddMute(ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.VOICE_UNMUTE_PLAYER_OPTION) then
			if (ChatmanagerDropDown.thisName) then
				DelMute(ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.VOICE_SILENCE_PLAYER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelSilenceVoice(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.VOICE_UNSILENCE_PLAYER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelUnSilenceVoice(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.MUTE_PLAYER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelMute(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.UNMUTE_PLAYER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelUnmute(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.KICK_PLAYER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelKick(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.BAN_PLAYER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelBan(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.SET_MODERATOR_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelModerator(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.UNSET_MODERATOR_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				ChannelUnmoderator(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		elseif (self.value == CHATMANAGER_TEXT.SET_OWNER_OPTION) then
			if (cdata and moderator and ChatmanagerDropDown.thisName) then
				SetChannelOwner(tostring(_selectedChannelIndex), ChatmanagerDropDown.thisName);
			end
		end
	end
end


function Chatmanager.Frame_DropdownCancel()
	CloseDropDownMenus();
end


function Chatmanager.Frame_SetPassword(password)
	-- Updates the password for the current channel.  If the player is not the
	-- owner or moderator, then the password is simply recorded.
	if (not password) then
		password = "";
	end
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Frame_SetPassword(\""..password.."\")");
	end
	if (_selectedChannelIndex) then
		local cdata = _channelList[_selectedChannelIndex];
		if (cdata and cdata.name and (cdata.type == 0)) then
			cdata.password = password;
			if (cdata.amOwner or cdata.amModerator) then
				_original_SetChannelPassword(tostring(_selectedChannelIndex), password);
			end
		end
	end
end


function Chatmanager.OptionsFrame_OnLoad()
	-- Make our frame closable with the ESC key
	tinsert(UISpecialFrames, "ChatmanagerOptionsFrame");

	-- Set text for titles,  buttons, tooltips, etc.
	ChatmanagerOptionsFrameTitleText:SetText(CHATMANAGER_TEXT.OPTIONS_FRAME_TITLE);
	ChatmanagerOptionsFrameReplyToSentText:SetText(CHATMANAGER_TEXT.REPLY_TO_SENT_TITLE);
	ChatmanagerOptionsFrameReplyToSentCheckButton.tooltipText = CHATMANAGER_TEXT.REPLY_TO_SENT_TOOLTIP;
	ChatmanagerOptionsFrameChatLinkingText:SetText(CHATMANAGER_TEXT.CHAT_LINK_TITLE);
	ChatmanagerOptionsFrameChatLinkingCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_LINK_TOOLTIP;
	ChatmanagerOptionsFrameSmartStickyText:SetText(CHATMANAGER_TEXT.SMART_STICKY_TITLE);
	ChatmanagerOptionsFrameSmartStickyCheckButton.tooltipText = CHATMANAGER_TEXT.SMART_STICKY_TOOLTIP;
	ChatmanagerOptionsFrameChatLogText:SetText(CHATMANAGER_TEXT.CHAT_LOG_TITLE);
	ChatmanagerOptionsFrameChatLogCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_LOG_TOOLTIP;
	ChatmanagerOptionsFrameShowButtonText:SetText(CHATMANAGER_TEXT.CHAT_BUTTON_TITLE);
	ChatmanagerOptionsFrameShowButtonCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_BUTTON_TOOLTIP;
	ChatmanagerOptionsFrameStickySayText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_SAY_TITLE);
	ChatmanagerOptionsFrameStickySayCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_SAY_TOOLTIP;
	ChatmanagerOptionsFrameStickyPartyText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_PARTY_TITLE);
	ChatmanagerOptionsFrameStickyPartyCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_PARTY_TOOLTIP;
	ChatmanagerOptionsFrameStickyRaidText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_RAID_TITLE);
	ChatmanagerOptionsFrameStickyRaidCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_RAID_TOOLTIP;
	ChatmanagerOptionsFrameStickyGuildText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_GUILD_TITLE);
	ChatmanagerOptionsFrameStickyGuildCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_GUILD_TOOLTIP;
	ChatmanagerOptionsFrameStickyOfficerText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_OFFICER_TITLE);
	ChatmanagerOptionsFrameStickyOfficerCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_OFFICER_TOOLTIP;
	ChatmanagerOptionsFrameStickyWhisperText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_WHISPER_TITLE);
	ChatmanagerOptionsFrameStickyWhisperCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_WHISPER_TOOLTIP;
	ChatmanagerOptionsFrameStickyBattlegroundText:SetText(CHATMANAGER_TEXT.CHAT_STICKY_BATTLEGROUND_TITLE);
	ChatmanagerOptionsFrameStickyBattlegroundCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_STICKY_BATTLEGROUND_TOOLTIP;
	ChatmanagerOptionsFrameScrollChatText:SetText(CHATMANAGER_TEXT.CHAT_SCROLL_TITLE);
	ChatmanagerOptionsFrameScrollChatCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_SCROLL_TOOLTIP;
	ChatmanagerOptionsFrameHideChatButtonsText:SetText(CHATMANAGER_TEXT.CHAT_HIDE_BUTTONS_TITLE);
	ChatmanagerOptionsFrameHideChatButtonsCheckButton.tooltipText = CHATMANAGER_TEXT.CHAT_HIDE_BUTTONS_TOOLTIP;
	ChatmanagerOptionsFrameUseGuildNotesText:SetText(CHATMANAGER_TEXT.USE_GUILD_NOTES_TITLE);
	ChatmanagerOptionsFrameUseGuildNotesCheckButton.tooltipText = CHATMANAGER_TEXT.USE_GUILD_NOTES_TOOLTIP;
	ChatmanagerOptionsFrameEditBoxAtTopText:SetText(CHATMANAGER_TEXT.EDITBOX_AT_TOP_TITLE);
	ChatmanagerOptionsFrameEditBoxAtTopCheckButton.tooltipText = CHATMANAGER_TEXT.EDITBOX_AT_TOP_TOOLTIP;
	ChatmanagerOptionsFrameEditBoxNoAltText:SetText(CHATMANAGER_TEXT.EDITBOX_NO_ALT_TITLE);
	ChatmanagerOptionsFrameEditBoxNoAltCheckButton.tooltipText = CHATMANAGER_TEXT.EDITBOX_NO_ALT_TOOLTIP;
	
	-- Handle locale specific sizes
	if (GetLocale() == "zhTW") then
		ChatmanagerOptionsFrame:SetWidth(300);
	end
end


function Chatmanager.OptionsFrame_OnShow()
	if (ChatmanagerState.Options.ReplyLastTell == 1) then
		ChatmanagerOptionsFrameReplyToSentCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameReplyToSentCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.ChatLink == 1) then
		ChatmanagerOptionsFrameChatLinkingCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameChatLinkingCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.SmartSticky == 1) then
		ChatmanagerOptionsFrameSmartStickyCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameSmartStickyCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.ChatLog == 1) then
		ChatmanagerOptionsFrameChatLogCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameChatLogCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.ShowButton == 1) then
		ChatmanagerOptionsFrameShowButtonCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameShowButtonCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.EnableChatScroll == 1) then
		ChatmanagerOptionsFrameScrollChatCheckButton:SetChecked(1);
		ChatmanagerOptionsFrameHideChatButtonsCheckButton:Enable();
		ChatmanagerOptionsFrameHideChatButtonsText:SetTextColor(_colorTextEnabled.r, _colorTextEnabled.g, _colorTextEnabled.b);
	else
		ChatmanagerOptionsFrameScrollChatCheckButton:SetChecked(0);
		ChatmanagerOptionsFrameHideChatButtonsCheckButton:Disable();
		ChatmanagerOptionsFrameHideChatButtonsText:SetTextColor(_colorTextDisabled.r, _colorTextDisabled.g, _colorTextDisabled.b);
	end
	if (ChatmanagerState.Options.HideChatButtons == 1) then
		ChatmanagerOptionsFrameHideChatButtonsCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameHideChatButtonsCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.UseGuildNote == 1) then
		ChatmanagerOptionsFrameUseGuildNotesCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameUseGuildNotesCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.EditBoxAtTop == 1) then
		ChatmanagerOptionsFrameEditBoxAtTopCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameEditBoxAtTopCheckButton:SetChecked(0);
	end
	if (ChatmanagerState.Options.EditBoxNoAlt == 1) then
		ChatmanagerOptionsFrameEditBoxNoAltCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameEditBoxNoAltCheckButton:SetChecked(0);
	end

	local color = ChatTypeInfo["SAY"];
	ChatmanagerOptionsFrameStickySayText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.SAY == 1) then
		ChatmanagerOptionsFrameStickySayCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickySayCheckButton:SetChecked(0);
	end
	color = ChatTypeInfo["PARTY"];
	ChatmanagerOptionsFrameStickyPartyText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.PARTY == 1) then
		ChatmanagerOptionsFrameStickyPartyCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickyPartyCheckButton:SetChecked(0);
	end
	color = ChatTypeInfo["RAID"];
	ChatmanagerOptionsFrameStickyRaidText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.RAID == 1) then
		ChatmanagerOptionsFrameStickyRaidCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickyRaidCheckButton:SetChecked(0);
	end
	color = ChatTypeInfo["GUILD"];
	ChatmanagerOptionsFrameStickyGuildText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.GUILD == 1) then
		ChatmanagerOptionsFrameStickyGuildCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickyGuildCheckButton:SetChecked(0);
	end
	color = ChatTypeInfo["OFFICER"];
	ChatmanagerOptionsFrameStickyOfficerText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.OFFICER == 1) then
		ChatmanagerOptionsFrameStickyOfficerCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickyOfficerCheckButton:SetChecked(0);
	end
	color = ChatTypeInfo["WHISPER"];
	ChatmanagerOptionsFrameStickyWhisperText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.WHISPER == 1) then
		ChatmanagerOptionsFrameStickyWhisperCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickyWhisperCheckButton:SetChecked(0);
	end
	color = ChatTypeInfo["BATTLEGROUND"];
	ChatmanagerOptionsFrameStickyBattlegroundText:SetTextColor(color.r, color.g, color.b);
	if (ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky.BATTLEGROUND == 1) then
		ChatmanagerOptionsFrameStickyBattlegroundCheckButton:SetChecked(1);
	else
		ChatmanagerOptionsFrameStickyBattlegroundCheckButton:SetChecked(0);
	end
end


function Chatmanager.OptionsFrame_ReplyToSentOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.ReplyLastTell = 1;
	else
		ChatmanagerState.Options.ReplyLastTell = 0;
	end
end


function Chatmanager.OptionsFrame_ChatLinkOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.ChatLink = 1;
	else
		ChatmanagerState.Options.ChatLink = 0;
	end
	Chatmanager.OptionsFrame_OnShow();
end


function Chatmanager.OptionsFrame_SmartStickyOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.SmartSticky = 1;
	else
		ChatmanagerState.Options.SmartSticky = 0;
	end
end


function Chatmanager.OptionsFrame_ChatLogOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.ChatLog = 1;
	else
		ChatmanagerState.Options.ChatLog = 0;
	end
end


function Chatmanager.OptionsFrame_ShowButtonOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.ShowButton = 1;
		ChatmanagerMenuButtonFrame:Show();
	else
		ChatmanagerState.Options.ShowButton = 0;
		ChatmanagerMenuButtonFrame:Hide();
	end
end


function Chatmanager.OptionsFrame_ScrollChatOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.EnableChatScroll = 1;
	else
		ChatmanagerState.Options.EnableChatScroll = 0;
	end
	Chatmanager.UpdateScrollingAndButtons();
	Chatmanager.OptionsFrame_OnShow();
end


function Chatmanager.OptionsFrame_HideChatButtonsOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.HideChatButtons = 1;
	else
		ChatmanagerState.Options.HideChatButtons = 0;
	end
	Chatmanager.UpdateScrollingAndButtons();
end


function Chatmanager.OptionsFrame_UseGuildNotesOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.UseGuildNote = 1;
	else
		ChatmanagerState.Options.UseGuildNote = 0;
	end
	Chatmanager.Frame_UpdatePlayerList();
end


function Chatmanager.OptionsFrame_EditBoxAtTopOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.EditBoxAtTop = 1;
	else
		ChatmanagerState.Options.EditBoxAtTop = 0;
	end
	Chatmanager.UpdateEditBoxPosition();
end


function Chatmanager.OptionsFrame_EditBoxNoAltOnClick(self)
	if (self:GetChecked()) then
		ChatmanagerState.Options.EditBoxNoAlt = 1;
	else
		ChatmanagerState.Options.EditBoxNoAlt = 0;
	end
	Chatmanager.UpdateEditBoxAttribs();
end


function Chatmanager.OptionsFrame_StickyTypeButtonOnClick(self, type)
	if (not type) then
		return;
	end
	if (self:GetChecked()) then
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky[type] = 1;
		ChatTypeInfo[type].sticky = 1;
	else
		ChatmanagerState.Servers[_serverName].Characters[_playerName].Sticky[type] = 0;
		ChatTypeInfo[type].sticky = 0;
	end
end


function Chatmanager.MenuButtonFrame_OnLoad()
	-- Set the tooltip text for the menu button
	ChatmanagerMenuButtonFrameMenuButton.tooltipText = CHATMANAGER_TEXT.FRAME_TITLE;
	ChatmanagerMenuButtonFrameMenuButton.tooltipText2 = CHATMANAGER_TEXT.MENU_BUTTON_DRAG_TOOLTIP;
end


function Chatmanager.Frame_ColorSwatchOnClick(self, index)
	-- Called for a specific client channel index, this function prepares and
	-- opens the color picker frame.  It can be called from a non-button
	-- context.
	if (index) then
		if (ColorPickerFrame:IsShown()) then
			if (ColorPickerFrame.cancelFunc) then
				ColorPickerFrame.cancelFunc(ColorPickerFrame.previousValues);
			end
			ColorPickerFrame:Hide();
		end
		local button = {};
		button.value = "CHANNEL"..index;
		button.r = ChatTypeInfo["CHANNEL"..index].r;
		button.g = ChatTypeInfo["CHANNEL"..index].g;
		button.b = ChatTypeInfo["CHANNEL"..index].b;
		button.swatchFunc = Chatmanager.Frame_SetChatColor;
		button.cancelFunc = Chatmanager.Frame_SetChatColor;
		UIDropDownMenuButton_OpenColorPicker(button);
	end
end


function Chatmanager.Frame_UpdateChannelVoice(id, enabled, active)
	-- Given a display line button id, this function updates its speaker
	-- frame.  See ChannelList_UpdateVoice for the source for this function.
	local speaker = getglobal("ChatmanagerChannelListFrameButton"..id.."SpeakerFrame");
	local speakerIcon =  getglobal("ChatmanagerChannelListFrameButton"..id.."SpeakerFrameOn");
	local speakerFlash = getglobal("ChatmanagerChannelListFrameButton"..id.."SpeakerFrameFlash");
	if (enabled) then
		if (active) then
			ChannelFrame_Desaturate(speakerIcon, nil, 1, 1, 1, 0.75);
			ChannelFrame_Desaturate(speakerFlash, nil, 1, 1, 1, 0.75);
		else
			ChannelFrame_Desaturate(speakerIcon, 1, nil, nil, nil, 0.5);
			ChannelFrame_Desaturate(speakerFlash, 1, nil, nil, nil, 0.5);
		end
		speakerFlash:Show();
		speaker:Show();
	else
		speaker:Hide();
	end
end


function Chatmanager.Frame_UpdatePlayerVoice(id, enabled, active, muted)
	-- Given a display line button id, this function updates its speaker
	-- frame.  See ChannelRoster_UpdateVoice for the source of this function.
	local speaker = getglobal("ChatmanagerPlayerListFrameButton"..id.."SpeakerFrame");
	local speakerIcon =  getglobal("ChatmanagerPlayerListFrameButton"..id.."SpeakerFrameOn");
	local speakerFlash = getglobal("ChatmanagerPlayerListFrameButton"..id.."SpeakerFrameFlash");
	local speakerMuted = getglobal("ChatmanagerPlayerListFrameButton"..id.."SpeakerFrameMuted");
	if (enabled) then
		if (active) then
			ChannelFrame_Desaturate(speakerIcon, nil, 1, 1, 1, 0.75);
			ChannelFrame_Desaturate(speakerFlash, nil, 1, 1, 1, 0.75);
			ChannelFrame_Desaturate(speakerMuted, nil, 1, 1, 1, 0.75);
		else
			ChannelFrame_Desaturate(speakerIcon, 1, nil, nil, nil, 0.5);
			ChannelFrame_Desaturate(speakerFlash, 1, nil, nil, nil, 0.5);
			ChannelFrame_Desaturate(speakerMuted, 1, nil, nil, nil, 0.75);
		end
		if (muted) then
			speakerMuted:Show();
		else
			speakerMuted:Hide();
		end
		speakerFlash:Show();
		speaker:Show();
	elseif (muted) then
		ChannelFrame_Desaturate(speakerIcon, 1, nil, nil, nil, 0.5);
		ChannelFrame_Desaturate(speakerMuted, 1, nil, nil, nil, 0.75);
		speakerFlash:Hide();
		speakerMuted:Show();
		speaker:Show();
	else
		speaker:Hide();
	end
end


function Chatmanager.Frame_HideSecureButtons()
	-- When entering combat we want the main Chatmanager window to be
	-- unencumbered by the usual restrictions for action buttons (can't show
	-- or hide them, etc.) which are propogated to their parents.  To do this,
	-- we remove them from the main window and hide them now.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Frame_HideSecureButtons()");
	end
	for i = 1, CHATMANAGER_PLAYER_LIST_BUTTON_COUNT do
		local actionButton = getglobal("ChatmanagerPlayerListFrameButton"..i.."Action");
		if (actionButton) then
			actionButton:SetParent(UIParent);
			actionButton:ClearAllPoints();
			actionButton:Hide();
		end
	end
end


function Chatmanager.Frame_ShowSecureButtons()
	-- Reset the action buttons back to their normal parent button
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.Frame_ShowSecureButtons()");
	end
	for i = 1, CHATMANAGER_PLAYER_LIST_BUTTON_COUNT do
		local button = getglobal("ChatmanagerPlayerListFrameButton"..i);
		local actionButton = getglobal("ChatmanagerPlayerListFrameButton"..i.."Action");
		if (button and actionButton) then
			actionButton:SetParent(button);
			actionButton:SetPoint("TOPLEFT", button, "TOPLEFT");
			actionButton:Show();
		end
	end
end


------------------------------------------------------------------------------
-- Popup support functions
------------------------------------------------------------------------------
function Chatmanager.GetPopupData(type)
	if (type == "PLAYERNOTE") then
		local note;
		if (ChatmanagerDropDown.thisName) then
			note = ChatmanagerState.Servers[_serverName].Notes[ChatmanagerDropDown.thisName];
		end
		return note or "";
	end
end


function Chatmanager.HandlePopupAccept(type, data, text)
	if (type == "LEAVE") then
		-- data is the channel name (index as a string) to leave
		if (data and (data ~= "")) then
			LeaveChannelByName(data);
		end
	elseif (type == "JOIN") then
		-- text is the contents of the edit box (name [password])
		local name = string.gsub(text, "%s*([^%s]+).*", "%1");
		local password = string.gsub(text, "%s*([^%s]+)%s*([^%s]*)", "%2");
		if (name and (string.len(name) > 0)) then
			local zoneID, channelName = Chatmanager.JoinPermanentChannel(name, password, DEFAULT_CHAT_FRAME:GetID(), 1);
			if (zoneID) then
				if (channelName) then
					name = channelName;
				end
				local sdata = _channelSavedList[string.lower(name)];
				if (sdata) then
					sdata.index = _selectedChannelDisplayIndex;
				end
				ChatFrame_RemoveChannel(DEFAULT_CHAT_FRAME, name);
				ChatFrame_AddChannel(DEFAULT_CHAT_FRAME, name);
			end
		end
	elseif (type == "PLAYERNOTE") then
		-- data is the player name, text is the new player note
		if (data) then
			if ((text == "") or (string.gsub(text, "%s", "") == "")) then
				text = nil;
			end
			local oldText = ChatmanagerState.Servers[_serverName].Notes[data];
			ChatmanagerState.Servers[_serverName].Notes[data] = text;
			if ((oldText ~= text) and DEFAULT_CHAT_FRAME) then
				if (text) then
					DEFAULT_CHAT_FRAME:AddMessage(string.format(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_CONFIRM_FORMAT, data, text));
				else
					if (Chatmanager.GetPlayerOrGuildNote(data)) then
						DEFAULT_CHAT_FRAME:AddMessage(string.format(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_REMOVE_DEFAULT_FORMAT, data));
					else
						DEFAULT_CHAT_FRAME:AddMessage(string.format(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_REMOVE_FORMAT, data));
					end
				end
			end
			Chatmanager.Frame_UpdatePlayerList();
		end
	end
end


------------------------------------------------------------------------------
-- Parsing functions
------------------------------------------------------------------------------
function Chatmanager.UpdateChannelList(expandAll, noUpdate)
	-- Uses the new GetChannelDisplayInfo function to make sure that we know
	-- about all the current channels (including the suspended ones which
	-- won't otherwise show up in the channel lists).
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.UpdateChannelList("..tostring(expandAll)..", "..tostring(noUpdate)..")");
	end
	if (_channelListUpdateIgnore or _channelListUpdateTimeout) then
		if (_debugFrame) then
			_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> doing nothing due to inhibit flags");
		end
		return;
	end
	for index, cdata in ipairs(_channelList) do
		cdata.channelInfoID = nil;
	end
	for id = 1, MAX_CHANNEL_BUTTONS do
		local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo(id);
		if (header) then
			if (collapsed and (not ChannelFrame:IsShown() or expandAll)) then
				-- Only expand the channel headers if the user isn't using the
				-- interface, where since this happens immediately we also
				-- have to make sure we don't call into ourselves again.
				_channelListUpdateIgnore = 1;
				ExpandChannelHeader(id);
				_channelListUpdateIgnore = nil;
			end
		elseif (name and channelNumber and (category ~= "CHANNEL_CATEGORY_GROUP")) then
			local index = tonumber(channelNumber);
			local cdata = _channelList[index or 0];
			if (cdata) then
				if (cdata.name == name) then
					cdata.channelInfoID = id;
					cdata.voiceEnabled = voiceEnabled;
					cdata.voiceActive = voiceActive;
				elseif (not active) then
					-- Only add this to the list if it is a suspended channel
					-- and we didn't already know about it
					local lname = string.lower(name);
					local _, fullName = _original_GetChannelName(tostring(index));
					cdata.name = name;
					cdata.fullName = fullName;
					_channelIndexLookupByName[lname] = index;
					if (category == "CHANNEL_CATEGORY_WORLD") then
						cdata.type = 1;
						if (not active) then
							cdata.suspended = true;
						end
					else
						cdata.type = 0;
					end
					-- Try and find an entry in the saved list to get the
					-- display index and other saved parameters
					local sdata = _channelSavedList[lname];
					if (sdata) then
						cdata.password = sdata.newPassword or sdata.password;
						cdata.isSticky = sdata.isSticky;
						cdata.forcePassword = sdata.forcePassword;
						if (sdata.index and not _channelIndexLookupByDisplayIndex[sdata.index]) then
							cdata.displayIndex = sdata.index;
							_channelIndexLookupByDisplayIndex[sdata.index] = index;
						end
						if (sdata.color) then
							ChangeChatColor("CHANNEL"..index, sdata.color.r, sdata.color.g, sdata.color.b);
						end
						sdata.newPassword = nil;
					end
					-- If we don't have a display index yet, see if we can
					-- assign its natural order index, otherwise assign it to
					-- the first free one.
					if (not cdata.displayIndex) then
						if (not _channelIndexLookupByDisplayIndex[index]) then
							cdata.displayIndex = index;
							_channelIndexLookupByDisplayIndex[index] = index;
						else
							for j = 1, CHATMANAGER_LAST_CHANNEL do
								if (not _channelIndexLookupByDisplayIndex[j]) then
									cdata.displayIndex = j;
									_channelIndexLookupByDisplayIndex[j] = index;
									break;
								end
							end
						end
					end
					-- Request the player list if this is a private channel
					if (cdata.type == 0) then
						Chatmanager.RequestPlayerList(index, false, false);
					end
					cdata.channelInfoID = id;
					cdata.voiceEnabled = voiceEnabled;
					cdata.voiceActive = voiceActive;
				else
					-- This is an active channel that we don't have in our
					-- list yet, which probably means CHANNEL_UI_UPDATE
					-- probably came before we got the YOU_JOINED message,
					-- so try again shortly
					_channelListUpdateTimeout = GetTime() + CHATMANAGER_CHANNEL_JOIN_UPDATE_DELAY;
					_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
				end
			end
		elseif (id > GetNumDisplayChannels()) then
			break;
		end
	end
	if (not noUpdate) then
		Chatmanager.Frame_UpdateChannelList();
	end
end


function Chatmanager.UpdateChannelListSelection()
	-- Determines what the current selected channel ID should be based on the
	-- channel display info and calls Chatmanager.Frame_UpdateChannelList to
	-- update this
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.UpdateChannelListSelection()");
	end
	local selectedID = GetSelectedDisplayChannel() or (GetActiveVoiceChannel and GetActiveVoiceChannel());
	local displayIndex;
	if (selectedID) then
		local name, header, _, channelNumber = GetChannelDisplayInfo(selectedID);
		if (name and not header and channelNumber) then
			local index = tonumber(channelNumber);
			local cdata = _channelList[index or 0];
			if (cdata and cdata.name) then
				displayIndex = cdata.displayIndex;
			end
		end
	end
	if (displayIndex) then
		Chatmanager.Frame_UpdateChannelList(displayIndex);
	else
		Chatmanager.Frame_UpdateChannelList("");
	end
end


function Chatmanager.RequestPlayerList(index, user, doRoster)
	-- With the WoW 2.2 channel changes we can now get player lists one of two
	-- ways - via GetChannelRosterInfo and via the previous player list
	-- mechanism.  Here we only use the old method (getting a player list) if
	-- doRoster is not set, or the user has explicitly requested it.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.RequestPlayerList("..tostring(index)..", "..tostring(user)..", "..tostring(doRoster)..")");
	end

	local cdata = _channelList[index];
	if (not cdata or not cdata.name) then
		return;
	end
	if (cdata.pending) then
		if (not cdata.request) then
			cdata.request = cdata.pending + CHATMANAGER_PLAYER_REQUEST_DELAY;
		end
		if (user) then
			cdata.nextUserRequest = true;
		end
	elseif (not cdata.request) then
		if (doRoster and cdata.channelInfoID) then
			Chatmanager.UpdateChannelRoster(cdata.channelInfoID, cdata);
		elseif (cdata.playerList) then
			for name, pdata in pairs(cdata.playerList) do
				pdata.pending = true;
			end
		else
			cdata.playerList = {};
		end
		cdata.pending = GetTime() + CHATMANAGER_PLAYER_UPDATE_INITIAL_DELAY;
		if (not _playerListUpdateTimeout) then
			_playerListUpdateTimeout = cdata.pending;
			_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
		end
		if (user) then
			cdata.userRequest = true;
		end
		if (user or not doRoster) then
			_original_ListChannelByName(tostring(index));
		end
	elseif (user) then
		cdata.nextUserRequest = true;
	end
end


function Chatmanager.UpdateChannelRoster(id, cdata)
	-- This function can be called in response to a CHANNEL_ROSTER_UPDATE
	-- event where the id parameter is the button ID in the Blizzard channel
	-- display list, or in response to another action where cdata will
	-- already be determined.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.UpdateChannelRoster("..tostring(id)..", "..tostring(cdata)..")");
	end
	if (not cdata) then
		for index, data in ipairs(_channelList) do
			if (data.channelInfoID == id) then
				cdata = data;
				break;
			end
		end
	end
	if (cdata and cdata.name and not cdata.suspended) then
		local _, header, _, _, count = GetChannelDisplayInfo(id);
		if (count and not header) then
			local playerList = cdata.playerList;
			if (playerList) then
				for name, pdata in pairs(playerList) do
					pdata.pending = true;
				end
			else
				playerList = {};
				cdata.playerList = playerList;
			end
			local voiceChat = IsVoiceChatEnabled();
			local noData, noUpdate;
			for index = 1, count do
				local name, isOwner, isModerator, voiceMuted, voiceActive, voiceEnabled = GetChannelRosterInfo(id, index);
				if (name) then
					local pdata = playerList[name];
					if (pdata) then
						pdata.pending = nil;
					else
						pdata = {};
						playerList[name] = pdata;
					end
					pdata.isOwner = isOwner;
					pdata.isModerator = isModerator;
					if (voiceChat) then
						pdata.voiceMuted = voiceMuted;
						pdata.voiceActive = voiceActive;
						pdata.voiceEnabled = voiceEnabled;
					end
				else
					-- GetChannelRosterInfo never seems to have only some data.
					-- It either has everything, or nothing, so skip the rest
					-- of the loop  if we don't get a name at any point
					noData = 1;
					break;
				end
			end
			for name, pdata in pairs(playerList) do
				if (pdata.pending) then
					if (noData) then
						-- If we didn't complete the roster, and there are pending
						-- entries, then set the timeout to clean the list up
						cdata.pending = GetTime() + CHATMANAGER_PLAYER_UPDATE_ROSTER_DELAY;
						if (not _playerListUpdateTimeout) then
							_playerListUpdateTimeout = cdata.pending;
							_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
						end
						noUpdate = 1;
						break;
					else
						playerList[name] = nil;
					end
				end
			end
			local pdata = playerList[_playerName];
			if (pdata and not pdata.pending) then
				if (pdata.isOwner) then
					cdata.amOwner = true;
					cdata.amModerator = nil;
				elseif (pdata.isModerator) then
					cdata.amOwner = nil;
					cdata.amModerator = true;
				else
					cdata.amOwner = nil;
					cdata.amModerator = nil;
				end
			end
			if (not noUpdate) then
				Chatmanager.RebuildPlayerFilteredList(cdata);
				Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
				if (_selectedChannelDisplayIndex and (_selectedChannelDisplayIndex == cdata.displayIndex)) then
					Chatmanager.Frame_UpdatePlayerList();
				end
			end
		end
	end
end


function Chatmanager.ProcessPlayerList(list, index)
	-- Processes messages resulting from a call to ListChannelByName.  Note
	-- that a single channel listing may be broken into multiple messages.  If
	-- the listing is for a user request, then the function returns true,
	-- otherwise it returns nil.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ProcessPlayerList("..list..", "..index..")");
	end
	local cdata = _channelList[index];
	if (cdata and cdata.name) then
		local pdata;
		if (not cdata.playerList) then
			cdata.playerList = {};
		end
		for qualifier, name in string.gmatch(list, CHATMANAGER_TEXT.LIST_PLAYER_NAMES_PARSE_STRING) do
			pdata = cdata.playerList[name];
			if (not pdata) then
				pdata = {};
				cdata.playerList[name] = pdata;
			else
				pdata.pending = nil;
			end
			if (qualifier == CHATMANAGER_TEXT.LIST_PLAYER_OWNER) then
				pdata.isOwner = true;
				pdata.isModerator = nil;
				pdata.isMuted = nil;
			elseif (qualifier == CHATMANAGER_TEXT.LIST_PLAYER_MODERATOR) then
				pdata.isOwner = nil;
				pdata.isModerator = true;
				pdata.isMuted = nil;
			elseif (qualifier == CHATMANAGER_TEXT.LIST_PLAYER_MUTED) then
				pdata.isOwner = nil;
				pdata.isModerator = nil;
				pdata.isMuted = true;
				cdata.isModerated = true;
			else
				pdata.isOwner = nil;
				pdata.isModerator = nil;
				pdata.isMuted = nil;
			end
		end
		pdata = cdata.playerList[_playerName];
		if (pdata) then
			if (pdata.isOwner) then
				cdata.amOwner = true;
				cdata.amModerator = nil;
			elseif (pdata.isModerator) then
				cdata.amOwner = nil;
				cdata.amModerator = true;
			else
				cdata.amOwner = nil;
				cdata.amModerator = nil;
			end
		else
			cdata.amOwner = nil;
			cdata.amModerator = nil;
		end
		Chatmanager.RebuildPlayerFilteredList(cdata);
		Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
		cdata.pending = GetTime() + CHATMANAGER_PLAYER_UPDATE_DELAY;
		if (_playerListUpdateTimeout and (_playerListUpdateTimeout > cdata.pending)) then
			_playerListUpdateTimeout = cdata.pending;
		end
		return cdata.userRequest;
	end
end


function Chatmanager.ConvertChannelHeader(header, index, channel)
	-- Returns the rebuilt channel header where the header contains the index
	-- that the player thinks this channel is on, rather than the one it is
	-- according to the client.  If the index is the same in both cases, the
	-- original header (passed in) is returned.  Note that for failure
	-- situations where the index is not found, we follow the same logic that
	-- adding a new channel would have followed to try and get a reasonable
	-- channel number.
	local newIndex = _channelList[index].displayIndex;
	if (not newIndex) then
		local lname = string.lower(string.gsub(channel, "%s%-%s.*", ""));
		local sdata = _channelSavedList[lname];
		if (sdata and sdata.index and not _channelIndexLookupByDisplayIndex[sdata.index]) then
			newIndex = sdata.index;
		end
		if (not newIndex) then
			for i = 1, CHATMANAGER_LAST_CHANNEL do
				if (not _channelIndexLookupByDisplayIndex[i]) then
					newIndex = i;
					break;
				end
			end
		end
	end
	if (newIndex and (newIndex ~= index)) then
		return string.format(CHATMANAGER_TEXT.CHANNEL_HEADER, newIndex, arg9);
	end
	return header;
end


function Chatmanager.RebuildList(text)
	-- Called to rebuild the output string of a ListChannels command
	local list = {};
	for index, name in string.gmatch(text, CHATMANAGER_TEXT.LIST_CHANNELS_PARSE_STRING) do
		local newIndex = _channelList[tonumber(index)].displayIndex;
		table.insert(list, newIndex or index);
	end
	table.sort(list);
	text = "";
	for _, displayIndex in ipairs(list) do
		local index = _channelIndexLookupByDisplayIndex[displayIndex or ""];
		local cdata = _channelList[index or ""];
		if (cdata and cdata.name) then
			text = text..string.format(CHATMANAGER_TEXT.LIST_CHANNELS_FORMAT_STRING, displayIndex, cdata.fullName);
		end
	end
	return text;
end


function Chatmanager.ProcessPlayerAction(name, index, action)
	-- Called when a player joins or leaves a channel, this only occurs if
	-- announcement is on for the channel (which is outside of our control).
	-- At any rate, this function simply adds or removes the indicated player
	-- from the list for the channel, where the list will be refreshed when
	-- the UI is shown or the selected channel changed anyway.  Since the
	-- join message doesn't tell us about the player, we also request an
	-- update of the player list for the channel.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ProcessPlayerAction("..name..", "..index..", "..action..")");
	end
	local cdata = _channelList[index];
	if (not cdata) then
		return;
	end
	if (action == "JOIN") then
		local playerList = cdata.playerList;
		if (not playerList) then
			playerList = {};
			cdata.playerList = playerList;
		end
		if (not playerList[name]) then
			playerList[name] = {};
		end
		if (cdata.isModerated) then
			playerList[name].isMuted = true;
		end
		Chatmanager.RebuildPlayerFilteredList(cdata);
		Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
		Chatmanager.RequestPlayerList(index, false, (index == _selectedChannelIndex));
	elseif (action == "LEAVE") then
		local playerList = cdata.playerList;
		if (playerList) then
			playerList[name] = nil;
			Chatmanager.RebuildPlayerFilteredList(cdata);
			Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
			if (index == _selectedChannelIndex) then
				Chatmanager.Frame_UpdatePlayerList();
			end
		end
	end
end


function Chatmanager.ProcessNotice(channel, index, zone, action)
	-- Called for join, leave, suspended, and changed notices on individual
	-- channels, and also for wrong passwords.  Suspended appears to be a
	-- special case for trade channels and amounts to the equivalent of
	-- "left", and is followed by a "joined" when you re-enter a trade zone.
	-- Also called for wrong-password notification.  This function returns
	-- -1 if an entry was removed, 1 if one was added, 0 if there was no
	-- change, and nil if the output should be suppressed.  Note that for
	-- other notice events (such as NOT_MODERATOR, NOT_MODERATED, MUTED,
	-- etc.) we simply return 0 so that the caller can pass the notification
	-- on.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ProcessNotice("..channel..", "..index..", "..zone..", "..action..")");
	end
	if (action == "YOU_LEFT") then
		local name = string.gsub(channel, "%s%-%s.*", "");
		local lname = string.lower(name);
		_channelIndexLookupByName[lname] = nil;
		if (_channelList[index].displayIndex) then
			_channelIndexLookupByDisplayIndex[_channelList[index].displayIndex] = nil;
		end
		if (_channelList[index].name == name) then
			-- Save details on the channel for rejoining again in this
			-- session.
			local sdata = {};
			sdata.name = name;
			sdata.index = _channelList[index].displayIndex;
			if (_channelList[index].password and (_channelList[index].password ~= "") and (_channelList[index].type == 0)) then
				sdata.password = _channelList[index].password;
			end
			sdata.isSticky = _channelList[index].isSticky;
			sdata.forcePassword = _channelList[index].forcePassword;
			local info = ChatTypeInfo["CHANNEL"..index];
			sdata.color = {};
			sdata.color.r = info.r;
			sdata.color.g = info.g;
			sdata.color.b = info.b;
			_channelSavedList[lname] = sdata;
		end
		_channelList[index] = {};
		Chatmanager.Frame_UpdateChannelList();
		Chatmanager.Frame_UpdatePlayerList();
		return -1;
	elseif (action == "YOU_JOINED") then
		local name = string.gsub(channel, "%s%-%s.*", "");
		if (not _channelList[index] or (_channelList[index].name ~= name) or _channelList[index].suspended) then
			local lname = string.lower(name);
			if (_channelList[index].displayIndex) then
				_channelIndexLookupByDisplayIndex[_channelList[index].displayIndex] = nil;
			end
			_channelList[index] = {};
			_channelList[index].name = name;
			_channelList[index].fullName = channel;
			if (zone == 0) then
				_channelList[index].type = 0;
			else
				_channelList[index].type = 1;
			end
			_channelIndexLookupByName[lname] = index;
			local sdata = _channelSavedList[lname];
			if (sdata) then
				if (not sdata.newPassword and sdata.password and (sdata.password ~= "")) then
					-- Joined this channel without using our password.  If we
					-- should force it then do that now, otherwise discard the
					-- password we had saved.
					if (sdata.forcePassword == 1) then
						if (_debugFrame) then
							_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> applying password to channel "..name);
						end
						_original_SetChannelPassword(tostring(index), sdata.password);
						_channelList[index].password = sdata.password;
					end
					sdata.password = nil;
				else
					_channelList[index].password = sdata.newPassword;
					sdata.newPassword = nil;
				end
				_channelList[index].isSticky = sdata.isSticky;
				_channelList[index].forcePassword = sdata.forcePassword;
				if (sdata.index and not _channelIndexLookupByDisplayIndex[sdata.index]) then
					_channelList[index].displayIndex = sdata.index;
					_channelIndexLookupByDisplayIndex[sdata.index] = index;
				end
				if (sdata.color) then
					local info = ChatTypeInfo["CHANNEL"..index];
					info.r = sdata.color.r;
					info.g = sdata.color.g;
					info.b = sdata.color.b;
				end
				if (sdata.tabs) then
					-- Restore tabs on which this channel was on, and then
					-- discard the tabs list, since if the user manually
					-- changes their assigments and then leaves/rejoins, we
					-- don't want to use the old list.
					for tab = 1, NUM_CHAT_WINDOWS do
						local frame = getglobal("ChatFrame"..tab);
						if (frame) then
							ChatFrame_RemoveChannel(frame, name);
							for i = 1, #sdata.tabs do
								if (sdata.tabs[i] == tab) then
									ChatFrame_AddChannel(frame, name);
									break;
								end
							end
						end
					end
					sdata.tabs = nil;
				end
			else
				-- Use default channel colors
				local info = ChatTypeInfo["CHANNEL"..index];
				info.r = _defaultChannelColor.r;
				info.g = _defaultChannelColor.g;
				info.b = _defaultChannelColor.b;
			end
			if (not _channelList[index].displayIndex) then
				for i = 1, CHATMANAGER_LAST_CHANNEL do
					if (not _channelIndexLookupByDisplayIndex[i]) then
						_channelList[index].displayIndex = i;
						_channelIndexLookupByDisplayIndex[i] = index;
						break;
					end
				end
			end
			if (_channelList[index].type == 0) then
				Chatmanager.RequestPlayerList(index, false, false);
			end
			Chatmanager.Frame_UpdateChannelList();
			return 1;
		else
			return 0;
		end
	elseif (action == "SUSPENDED") then
		-- Note that only server channels can be suspended, so save the
		-- minimal data
		local name = string.gsub(channel, "%s%-%s.*", "");
		local lname = string.lower(name);
		local sdata = {};
		sdata.name = name;
		sdata.isSticky = _channelList[index].isSticky;
		local info = ChatTypeInfo["CHANNEL"..index];
		sdata.color = {};
		sdata.color.r = info.r;
		sdata.color.g = info.g;
		sdata.color.b = info.b;
		_channelSavedList[lname] = sdata;
		-- Update channel status
		_channelList[index].suspended = true;
		_channelList[index].fullName = name;
		_channelList[index].playerList = {};
		_channelList[index].filteredList = {};
		_channelList[index].playerCount = 0;
		Chatmanager.Frame_UpdateChannelList();
		Chatmanager.Frame_UpdatePlayerList();
		return 0;
	elseif (action == "YOU_CHANGED") then
		_channelList[index].fullName = channel;
		_channelList[index].playerList = {};
		_channelList[index].filteredList = {};
		_channelList[index].playerCount = 0;
		if (_channelList[index].type == 0) then
			Chatmanager.RequestPlayerList(index, false, (index == _selectedChannelIndex));
		end
		Chatmanager.Frame_UpdateChannelList();
		Chatmanager.Frame_UpdatePlayerList();
		return 0;
	elseif (action == "WRONG_PASSWORD") then
		-- When this happens, the client can get a bit confused about whether
		-- it is or isn't joined to the channel, so its best to force it to
		-- leave the channel explicitly, but note that in this case the
		-- "YOU_LEFT" message won't contain a valid index, so we won't see it
		-- here in this function and instead we do that cleanup now.  If there
		-- was a saved password then we return nil to indicate that we don't
		-- want to report this yet (since we will reattempt the connection
		-- when the empty YOU_LEFT notification is received).
		local lname = string.lower(string.gsub(channel, "%s%-%s.*", ""));
		_channelIndexLookupByName[lname] = nil;
		if (_channelList[index].displayIndex) then
			_channelIndexLookupByDisplayIndex[_channelList[index].displayIndex] = nil;
		end
		_channelList[index] = {};
		local sdata = _channelSavedList[lname];
		if (sdata and sdata.password) then
			return nil;
		end
		Chatmanager.Frame_UpdateChannelList();
		Chatmanager.Frame_UpdatePlayerList();
		return 0;
	end
	-- For unhandled actions, just return 0
	return 0;
end


function Chatmanager.ProcessEmptyNotice(channel, action)
	-- Called for actions that don't have an associated channel number and
	-- name in arg8 and arg9, this function is called with the channel name
	-- from arg4 instead.  Observed actions are YOU_LEFT and YOU_JOINED when
	-- a channel connection has failed (typically due to incorrect password).
	-- For YOU_LEFT actions, if there is still a password in the saved list
	-- then we attempt to connect again using it.  For YOU_JOINED actions we
	-- simply issue a disconnect.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ProcessEmptyNotice("..channel..", "..action..")");
	end
	if (action == "YOU_LEFT") then
		local lname = string.lower(string.gsub(channel, "%s%-%s.*", ""));
		local sdata = _channelSavedList[lname];
		if (sdata and sdata.password and (sdata.password ~= "")) then
			if (not _channelConnects) then
				_channelConnects = {};
			end
			_channelConnects[lname] = sdata;
			_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
		end
	elseif (action == "YOU_JOINED") then
		LeaveChannelByName(channel);
	end
end


function Chatmanager.ProcessUserNotice(name, index, action)
	-- Called for CHAT_MSG_NOTICE_USER events
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ProcessUserNotice("..name..", "..index..", "..action..")");
	end
	local cdata = _channelList[index];
	if (not cdata) then
		return;
	end
	local pdata = nil;
	if (name and cdata.playerList) then
		pdata = cdata.playerList[name];
	end
	if (action == "SET_VOICE") then
		if (pdata) then
			pdata.isMuted = nil;
		end
		cdata.isModerated = true;
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
	elseif (action == "UNSET_VOICE") then
		if (pdata) then
			pdata.isMuted = true;
		end
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
	elseif (action == "SET_MODERATOR") then
		if (pdata) then
			pdata.isModerator = true;
		end
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
	elseif (action == "UNSET_MODERATOR") then
		if (pdata) then
			pdata.isModerator = nil;
		end
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
	elseif (action == "MODERATION_ON") then
		cdata.isModerated = true;
		Chatmanager.RequestPlayerList(index, false, (index == _selectedChannelIndex));
	elseif (action == "MODERATION_OFF") then
		cdata.isModerated = nil;
		Chatmanager.RequestPlayerList(index, false, (index == _selectedChannelIndex));
	elseif (action == "PASSWORD_CHANGED") then
		-- Password changed, but we don't know what it is yet (unless of
		-- course we changed it).  For now, leave our own password alone.
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdateChannelList(_selectedChannelDisplayIndex);
		end
	elseif (action == "OWNER_CHANGED") then
		cdata.isModerated = nil;
		Chatmanager.RequestPlayerList(index, false, (index == _selectedChannelIndex));
	elseif (action == "PLAYER_BANNED") then
		-- Nothing special to do here, ban will be followed by a kick
		-- automatically
	elseif (action == "PLAYER_KICKED") then
		if (pdata) then
			cdata.playerList[name] = nil;
			Chatmanager.RebuildPlayerFilteredList(cdata);
			Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
		end
		cdata.isModerated = true;
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdatePlayerList();
		end
	end
end


function Chatmanager.ProcessPasswordRequest(channel)
	local lname = string.lower(string.gsub(channel, "%s%-%s.*", ""));
	local sdata = _channelSavedList[lname];
	if (sdata and sdata.password and (sdata.password ~= "")) then
		if (not _channelConnects) then
			_channelConnects = {};
		end
		_channelConnects[lname] = sdata;
		_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
		StaticPopup_Hide("CHAT_CHANNEL_PASSWORD");
	end

end


function Chatmanager.ConvertHyperLinks(text)
	-- Converts any hyperlinks in the passed text into a format compatible
	-- with Yrys's ChatLink so that we can carry this information on private
	-- chat channels.
	if (text) then	
		local newText, changes = string.gsub(text, "|c(%x+)|H(%a+):(.-)|h%[(.-)%]|h|r", "{CLINK:%2:%1:%3:%4}");
		if (changes > 0) then
			if (_debugFrame) then
				_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."converted \""..text.."\" (length "..string.len(text)..") to \""..newText.."\" (length "..string.len(newText)..")");
			end
			return newText, true;
		end
	end
end


function Chatmanager.RestoreLink(type, color, params)
	-- Used by the string.gsub call in Chatmanager.RestoreHyperLinks to do the
	-- reassembly of the parameters into the hyperlink.  This is pretty
	-- straight forward except for extracting the name from the params (which
	-- depends on the link type) and for quest links for which we need to
	-- adjust the link color based on the difference between the quest and
	-- player's levels.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.RestoreLink("..tostring(type or "nil")..", "..tostring(color or "nil")..", "..tostring(params or "nil")..")");
	end
	local name;
	if (type == "item") then
		params, name = string.match(params, "^(%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+):(.-)$");
	elseif ((type == "enchant") or (type == "spell")) then
		params, name = string.match(params, "^(%-?%d+):(.-)$");
	elseif (type == "quest") then
		local id, level, questColor;
		id, level, name = string.match(params, "^(%-?%d+):(%-?%d+):(.+)$");
		if (tonumber(level) > 0) then
			questColor = GetQuestDifficultyColor(tonumber(level));
		else
			questColor = QuestDifficultyColor["difficult"];
		end
		color = string.format("ff%02x%02x%02x", 255*questColor.r, 255*questColor.g, 255*questColor.b);
	elseif (type == "talent") then
		params, name = string.match(params, "^(%-?%d+:%-?%d+):(.-)$");
	elseif (type == "achievement") then
		params, name = string.match(params, "^(%-?%d+:%x+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+):(.-)$");
	elseif (type == "trade") then
		params, name = string.match(params, "^(%-?%d+:%-?%d+:%-?%d+:%x+:[^:]+):(.-)$");
	elseif (tonumber(color)) then
		-- Looks like it an item link from ChatLink without the "item" type
		params, name = string.match(color..":"..params, "^(%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+):(.-)$");
		color = type;
		type = "item";
	else
		-- Do best guess for future types of hyper links...
		params, name = string.match(params, "^(.+):([^:]+)$");
	end
	local result = "|c"..color.."|H"..type..":"..params.."|h["..name.."]|h|r";
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> \""..result.."\" (length "..string.len(result)..")");
	end
	return result;
end

function Chatmanager.RestoreHyperLinks(text)
	-- Converts any ChatLink style text back into the original hyperlink.
	-- This is pretty straight forward except for quest links, which are
	-- handled here in a manner similar to how Yrys does them (i.e., extract
	-- the quest level and then set the color based on the current player's
	-- level, using Blizzard's GetDifficultyColor function).
	if (text) then
		local newText = string.gsub(text, "{CLINK:([%a%d]+):(%x+):([^}]+)}", Chatmanager.RestoreLink);
		return newText;
	end
	return text;
end


------------------------------------------------------------------------------
-- Hooked functions
------------------------------------------------------------------------------
function Chatmanager.JoinPermanentChannel(name, password, ID, flag)
	-- Hooked so we can capture the password for the new channel
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.JoinPermanentChannel("..tostring(name)..", "..tostring(password)..", "..tostring(ID)..", "..tostring(flag)..")");
	end
	local index = _channelIndexLookupByName[string.lower(name)];
	if (index) then
		if (_debugFrame) then
			_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> already connected to channel \""..name.."\"");
		end
	else
		local zoneID, channelName = _original_JoinPermanentChannel(name, password, ID, flag);
		if (zoneID) then
			if (channelName) then
				name = channelName;
			end
			if (_debugFrame) then
				_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> zoneID = "..zoneID.." and name = "..name);
			end
			local lname = string.lower(name);
			if (not _channelSavedList[lname]) then
				-- Its a new channel, so create an entry and give it the
				-- default colors
				_channelSavedList[lname] = {};
				_channelSavedList[lname].name = name;
				_channelSavedList[lname].color = {};
				_channelSavedList[lname].color.r = _defaultChannelColor.r;
				_channelSavedList[lname].color.g = _defaultChannelColor.g;
				_channelSavedList[lname].color.b = _defaultChannelColor.b;
			end
			if (password and (password ~= "")) then
				_channelSavedList[lname].newPassword = password;
			else
				_channelSavedList[lname].newPassword = nil;
			end
			_channelSavedList[lname].password = nil;
			_channelListUpdateTimeout = GetTime() + CHATMANAGER_CHANNEL_JOIN_UPDATE_DELAY;
			_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
			return zoneID, channelName;
		end
	end
end


function Chatmanager.ListChannelByName(name, silent)
	-- Hooked so we can determine when some other component is requesting a
	-- channel listing.  It is assumed that the passed name is either the
	-- channel name or its actual index (not its display index).
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ListChannelByName("..name..")");
	end
	local index = tonumber(name);
	if (not index) then
		index = _channelIndexLookupByName[string.lower(name or "")];
	end
	if (index) then
		if (silent) then
			Chatmanager.RequestPlayerList(index, false, false);
		else
			Chatmanager.RequestPlayerList(index, true, false);
		end
	else
		_original_ListChannelByName(name);
	end
end


function Chatmanager.SlashCommand_LEAVE(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
		if (index) then
			LeaveChannelByName(tostring(index));
		end
	else
		LeaveChannelByName(name);
	end
end


function Chatmanager.SlashCommand_LIST_CHANNEL(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	if (string.len(name) > 0) then
		local index = tonumber(name);
		if (index) then
			index = _channelIndexLookupByDisplayIndex[index];
		else
			index = _channelIndexLookupByName[string.lower(name)];
		end
		if (index) then
			Chatmanager.RequestPlayerList(index, true, false);
		end
	else
		ListChannels();
	end
end


function Chatmanager.SlashCommand_CHAT_PASSWORD(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local password = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index) then
		_channelList[index].password = password;
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdateChannelList(_selectedChannelDisplayIndex);
		end
		_original_SetChannelPassword(tostring(index), password);
	else
		_original_SetChannelPassword(name, password);
	end
end


function Chatmanager.SlashCommand_CHAT_OWNER(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local newOwner = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index) then
		if (string.len(newOwner) > 0) then
			SetChannelOwner(tostring(index), newOwner);
		else
			DisplayChannelOwner(tostring(index));
		end
	end
end


function Chatmanager.SlashCommand_CHAT_MODERATOR(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelModerator(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_UNMODERATOR(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelUnmoderator(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_MODERATE(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index) then
		ChannelModerate(tostring(index));
	end
end


function Chatmanager.SlashCommand_CHAT_MUTE(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelMute(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_UNMUTE(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelUnmute(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_CINVITE(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelInvite(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_KICK(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelKick(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_BAN(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelBan(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_UNBAN(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local player = gsub(msg, "%s*([^%s]+)%s*(.*)", "%2");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index and (string.len(player) > 0)) then
		ChannelUnban(tostring(index), player);
	end
end


function Chatmanager.SlashCommand_CHAT_ANNOUNCE(msg)
	local name = gsub(msg, "%s*([^%s]+).*", "%1");
	local index = tonumber(name);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(name)];
	end
	if (index) then
		ChannelToggleAnnouncements(tostring(index));
	end
end


function Chatmanager.GetChannelName(channel)
	-- Hooked so that we can perform the channel number reassignment without
	-- actually having to leave and rejoin etc., and hope to get the channel
	-- numbers right, etc.  For the default client, GetChannelName is called
	-- only by the ChatFrame code.  Note that we want this function to return
	-- the actual channel number, etc., its only when its given our "display"
	-- channel number that we want to convert it.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.GetChannelName("..channel..")");
	end
	if (_convertChannelNumber) then
		local index = tonumber(channel or "");
		if (index) then
			local newIndex = _channelIndexLookupByDisplayIndex[index];
			if (newIndex) then
				channel = tostring(newIndex);
			else
				channel = nil;
			end
		end
	end
	if (channel and (channel ~= "")) then
		local num, name, id = _original_GetChannelName(channel);
		if (_debugFrame) then
			if (num and name) then
				if (id) then
					_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> "..num..", "..name..", "..id);
				else
					_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."--> "..num..", "..name);
				end
			end
		end
		return num, name, id;
	end
	return 0;
end


function Chatmanager.GetChannelList()
	-- Hooked so we can re-order the list according to what the user has
	-- selected.  We do this because we wish to maintain the correct color
	-- index association with the client, so that channel4 is really always 
	-- channel4, even though, via this addon we make it (say) channel7.  Note
	-- that the caller of GetChannelList expects each of the return arguments
	-- to be passed, not a single table, so we use unpack to do the conversion
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.GetChannelList()");
	end
	local list = {_original_GetChannelList()};
	local tempList = {};
	-- First reorder list into table based on display index (not best way to
	-- do this, but with a maximum of only 20 items in the table, we don't
	-- need to be super efficient).
	local maxPos = 0;
	for i = 1, #list, 2 do
		local index = list[i];
		local name = list[i+1];
		local newIndex = _channelList[index].displayIndex;
		if (newIndex) then
			local j = (newIndex*2) - 1;
			tempList[j] = index;
			tempList[j+1] = name;
			if (j + 1 > maxPos) then
				maxPos = j + 1;
			end
		else
			tempList[i] = index;
			tempList[i+1] = name;
			if (i + 1 > maxPos) then
				maxPos = i + 1;
			end
		end
	end
	-- Then compress to remove any sequence gaps
	local newList = {};
	for i = 1, maxPos, 2 do
		if (tempList[i] and tempList[i + 1]) then
			table.insert(newList, tempList[i]);
			table.insert(newList, tempList[i + 1]);
		end
	end
	return unpack(newList);
end


function Chatmanager.ChatEdit_UpdateHeader(editBox)
	-- Hooked so that we can clear the _convertChannelNumber flag if needed
	-- and to rebuild the header for chat channels where the display index is
	-- different to the actual index.  The implementation of this function is
	-- based on the original ChatEdit_UpdateHeader function (ChatFrame.lua).
--	if (_debugFrame) then
--		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatEdit_UpdateHeader("..editBox:GetName()..")");
--	end
	if (editBox:GetAttribute("chatType") == "CHANNEL") then
		_convertChannelNumber = nil;
		_original_ChatEdit_UpdateHeader(editBox);
		local header = getglobal(editBox:GetName().."Header");
		if (header) then
			-- editBox.channelTarget should have the correct (client) index
			local index = tonumber(editBox:GetAttribute("channelTarget"));
			local cdata = _channelList[index or 0];
			if (cdata and cdata.name and cdata.displayIndex) then
				local newIndex = _channelList[index].displayIndex;
				if (newIndex ~= index) then
					local text = string.gsub(header:GetText(), "%d+", tostring(newIndex), 1);
					if (text) then
						header:SetText(text);
						editBox:SetTextInsets(15 + header:GetWidth(), 13, 0, 0);
					end
				end
			end
		end
	else
		_original_ChatEdit_UpdateHeader(editBox);
	end
end


function Chatmanager.ChatFrame_ConfigEventHandler(self, event, ...)
	-- Hooked to enable us to unregister chat message events for unused chat
	-- windows.
	local result = _original_ChatFrame_ConfigEventHandler(self, event, ...);
	if (event == "UPDATE_CHAT_WINDOWS") then
		local _, _, _, _, _, _, shown, _, docked = GetChatWindowInfo(self:GetID());
		if (shown or docked) then
			self:RegisterEvent("CHAT_MSG_CHANNEL");
		else
			self:UnregisterEvent("CHAT_MSG_CHANNEL");
		end
	end
	return result;
end


function Chatmanager.ChatFrame_MessageEventHandler(self, event, ...)
	-- Hooked to enable us to intercept and discard or change channel related
	-- messages - particularly to support number reassignment and whispers.
	local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 = ...;
--	if (_debugFrame) then
--		if (event) then
--			_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatFrame_MessageEventHandler("..(self and self.GetName and self:GetName() or "nil")..", "..event..", "..(arg1 or "")..")");
--		else
--			_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatFrame_MessageEventHandler("..(self and self.GetName and self:GetName() or "nil")..")");
--		end
--	end
	if (string.sub(event, 1, 16) == "CHAT_MSG_CHANNEL") then
		if (event == "CHAT_MSG_CHANNEL") then
			if (arg9 and (arg9 ~= "") and arg8 and (arg8 ~= "")) then
				local index = tonumber(arg8);
				arg4 = Chatmanager.ConvertChannelHeader(arg4, index, arg9);
				local cdata = _channelList[index];
				if (cdata and cdata.name and (cdata.type == 0)) then
					-- This is a private chat channel, so check if there is
					-- item or enchant data in the message in Chatmanager or
					-- ChatLink format and convert them to actual links if so
					arg1 = Chatmanager.RestoreHyperLinks(arg1);
				end
			end
			_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
		elseif (event == "CHAT_MSG_CHANNEL_LIST") then
			-- List of all channels or a list of players on a channel
			if (arg9 and (arg9 ~= "") and arg8 and (arg8 ~= "")) then
				local index = tonumber(arg8);
				if (Chatmanager.ProcessPlayerList(arg1, index)) then
					-- This was an update explicitly requested by the user
					arg4 = Chatmanager.ConvertChannelHeader(arg4, index, arg9);
				else
					-- Don't pass this list request on
					return;
				end
			else
				-- List of all subscribed channnels, just need to rebuild this
				arg1 = Chatmanager.RebuildList(arg1);
			end
			_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
		elseif (event == "CHAT_MSG_CHANNEL_JOIN") then
			-- Someone has joined this chat channel
			Chatmanager.ProcessPlayerAction(arg2, tonumber(arg8), "JOIN");
			arg4 = Chatmanager.ConvertChannelHeader(arg4, tonumber(arg8), arg9);
			if (ChatmanagerState.Options.ShowPlayerNote == 1) then
				local note = Chatmanager.GetPlayerOrGuildNote(arg2);
				if (note and (note ~= "")) then
					-- If we have a note for this player, then add it to the
					-- string used to format the joined message by modifying
					-- the global variable that is used to show this.  Doing
					-- it this way is  a bit of a kludge, but is minimally
					-- invasive and it does work.
					local originalFormat = CHAT_CHANNEL_JOIN_GET;
					CHAT_CHANNEL_JOIN_GET = string.format(CHATMANAGER_TEXT.CHANNEL_JOIN_NOTE_FORMAT, CHAT_CHANNEL_JOIN_GET, note);
					_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
					CHAT_CHANNEL_JOIN_GET = originalFormat;
				else
					_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
				end
			end
		elseif (event == "CHAT_MSG_CHANNEL_LEAVE") then
			-- Someone has joined this chat channel
			Chatmanager.ProcessPlayerAction(arg2, tonumber(arg8), "LEAVE");
			arg4 = Chatmanager.ConvertChannelHeader(arg4, tonumber(arg8), arg9);
			_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
		elseif (event == "CHAT_MSG_CHANNEL_NOTICE") then
			-- Notice actions are joined, left, suspended, changed, and wrong
			-- password.  Because these affect the channel list contents we
			-- have to be careful about how we reformat arg4.
			if (arg9 and (arg9 ~= "") and arg8 and (arg8 ~= "")) then
				local index = tonumber(arg8);
				local newarg4 = Chatmanager.ConvertChannelHeader(arg4, index, arg9);
				local result = Chatmanager.ProcessNotice(arg9, index, arg7, arg1);
				if (result == 1) then
					arg4 = Chatmanager.ConvertChannelHeader(arg4, index, arg9);
					_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
				elseif (result) then
					arg4 = newarg4;
					_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
				end
			else
				Chatmanager.ProcessEmptyNotice(arg4, arg1);
			end
		elseif (event == "CHAT_MSG_CHANNEL_NOTICE_USER") then
			if (arg9 and (arg9 ~= "") and arg8 and (arg8 ~= "")) then
				local index = tonumber(arg8);
				arg4 = Chatmanager.ConvertChannelHeader(arg4, index, arg9);
				Chatmanager.ProcessUserNotice(arg2, index, arg1);
			end
			_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
		else
			if (arg9 and (arg9 ~= "") and arg8 and (arg8 ~= "")) then
				arg4 = Chatmanager.ConvertChannelHeader(arg4, tonumber(arg8), arg9);
			end
			_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
		end
	elseif (event == "CHAT_MSG_WHISPER_INFORM") then
		if (ChatmanagerState.Options.ReplyLastTell == 1) then
			if (self and self.editBox) then
				ChatEdit_SetLastTellTarget(arg2);
			end
		end
		_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
	elseif ((event == "CHAT_MSG_SYSTEM") and (ChatmanagerState.Options.ShowPlayerNote == 1)) then
		-- Check whether it is a "player has come online" message
		local _, _, player = string.find(arg1, _playerIsOnlineSearchPattern);
		if (player) then
			local note = Chatmanager.GetPlayerOrGuildNote(player);
			if (note and (note ~= "")) then
				arg1 = string.format(CHATMANAGER_TEXT.CHANNEL_JOIN_NOTE_FORMAT, arg1, note);
			end
		end
		_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
	else
		_original_ChatFrame_MessageEventHandler(self, event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, select(10, ...));
	end
	return true;
end


function Chatmanager.ChatEdit_HandleChatType(editBox, msg, command, send)
	-- Hooked so that we can set the _convertChannelNumber flag so if
	-- ChatEdit_HandleChatType calls our hooked GetChannelName function it
	-- will translate a user input channel number to the actual channel
	-- number.  (This flag will be cleared automatically by
	-- Chatmanager.ChatEdit_UpdateHeader if it is also called).  Note that
	-- since Blizzard's ChatEdit_ParseText calls this function via securecall,
	-- we don't need to worry about our hooks preventing execution of slash
	-- commands.
--	if (_debugFrame) then
--		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatEdit_HandleChatType("..editBox:GetName()..", "..tostring(msg or "")..", "..tostring(command or "")..", "..tostring(send or "")..")");
--	end
	_convertChannelNumber = true;
	local result = _original_ChatEdit_HandleChatType(editBox, msg, command, send);
	_convertChannelNumber = nil;
	return result;
end


function Chatmanager.ChatEdit_OnEscapePressed(editBox)
	-- Hooked to restore the previous sticky channel (if there was one) when
	-- cancelling an edit box operation, and to clear _channelOriginalText
--	if (_debugFrame) then
--		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatEdit_OnEscapePressed("..editBox:GetName()..")");
--	end
	_original_ChatEdit_OnEscapePressed(editBox);
	if ((editBox:GetAttribute("chatType") == "CHANNEL") and editBox.stickyChannel) then
		editBox:SetAttribute("channelTarget", editBox.stickyChannel);
	end
	_channelOriginalText = nil;
end


function Chatmanager.ChatEdit_AddHistory(editBox)
	-- Modified to store the display index for the channel rather than the
	-- actual index, and the original text rather than any item linking
	-- modified text.
--	if (_debugFrame) then
--		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatEdit_AddHistory("..editBox:GetName()..")");
--	end
	if (editBox:GetAttribute("chatType") == "CHANNEL") then
		local index = tonumber(editBox:GetAttribute("channelTarget"));
		local cdata = _channelList[index or 0];
		if (cdata and cdata.name and cdata.displayIndex) then
			editBox:SetAttribute("channelTarget", tostring(cdata.displayIndex));
			if (_channelOriginalText) then
				editBox:SetText(_channelOriginalText);
			end
			_original_ChatEdit_AddHistory(editBox);
			editBox:SetAttribute("channelTarget", tostring(index));
			_channelOriginalText = nil;
		else
			_original_ChatEdit_AddHistory(editBox);
		end
	else
		_original_ChatEdit_AddHistory(editBox);
	end
end


function Chatmanager.SendChatMessage(text, type, language, channel)
	-- Hooked to allow us to do the item linking in channel messages
--	if (_debugFrame) then
--		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.SendChatMessage("..tostring(text or "")..", "..tostring(type or "")..", "..tostring(language or "")..","..tostring(channel or "")..")");
--	end
	if (type == "CHANNEL") then
		local converted = nil;
		if (ChatmanagerState.Options.ChatLink == 1) then
			local index = tonumber(channel);
			local cdata = _channelList[index or 0];
			if (cdata and (cdata.type == 0)) then
				if (_original_FLT_ParseChatMessage) then
					text = _original_FLT_ParseChatMessage(text);
					converted = true;
				end
				local newText, changed = Chatmanager.ConvertHyperLinks(text);
				if (changed) then
					_channelOriginalText = text;
					text = newText;
				else
					_channelOriginalText = nil; 
				end
			end
		end
		_ignoreLinkConversions = converted;
		_original_SendChatMessage(text, type, language, channel);
		_ignoreLinkConversions = nil;
	else
		_original_SendChatMessage(text, type, language, channel);
	end
end


function Chatmanager.FCF_SetButtonSide(chatFrame, side)
	_original_FCF_SetButtonSide(chatFrame, side);
	if (_hideChatButtons == 1) then
		local bottomButton = getglobal(chatFrame:GetName().."BottomButton");
		if (bottomButton) then
			bottomButton:ClearAllPoints();
			bottomButton:SetPoint("BOTTOMRIGHT", chatFrame, "BOTTOMRIGHT", 0, -4);
		end
	end
end


function Chatmanager.SetChannelPassword(channel, password)
	-- Hooked so that we can see password changes that are made through the
	-- new 2.2 channels list interface
	local index = tonumber(channel);
	if (index) then
		index = _channelIndexLookupByDisplayIndex[index];
	else
		index = _channelIndexLookupByName[string.lower(channel)];
	end
	if (index) then
		_channelList[index].password = password;
		if (index == _selectedChannelIndex) then
			Chatmanager.Frame_UpdateChannelList(_selectedChannelDisplayIndex);
		end
		_original_SetChannelPassword(tostring(index), password);
	else
		_original_SetChannelPassword(channel, password);
	end
end


function Chatmanager.ChannelList_UpdateVoice(id, enabled, active)
	-- The original ChannelList_UpdateVoice takes the passed channel ID as the
	-- button ID, so we just need to change that to the actual ID of the
	-- button having the channel ID passed in.  Since this function may also
	-- be called within Chatmanager.ChannelList_Update, we check that first.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChannelList_UpdateVoice("..tostring(id)..", "..tostring(enabled)..", "..tostring(active)..")");
	end
	if (_channelListUpdateInProgress) then
		_original_ChannelList_UpdateVoice(id, enabled, active);
	else
		if (ChannelFrame:IsShown()) then
			local buttonIndex = _channelListButtonLookupById[id or 0];
			if (buttonIndex) then
				_original_ChannelList_UpdateVoice(buttonIndex, enabled, active);
			else
				_original_ChannelList_UpdateVoice(id, enabled, active);
			end
		end
		local name, header, _, channelNumber = GetChannelDisplayInfo(id);
		if (name and not header and channelNumber) then
			local index = tonumber(channelNumber);
			local cdata = _channelList[index or 0];
			if (cdata and cdata.name) then
				cdata.voiceEnabled = enabled;
				cdata.voiceActive = active;
				if (cdata.displayIndex) then
					Chatmanager.Frame_UpdateChannelVoice(cdata.displayIndex, enabled, active);
				end
			end
		end
	end
end


------------------------------------------------------------------------------
-- Post-hooked functions
------------------------------------------------------------------------------
function Chatmanager.ChatEdit_SendText(editBox, addHistory)
	-- Called in a secure manner via hooksecurefunc after ChatEdit_SendText
	-- has been processed, this function allows us to update the sticky state
	-- for the editBox.  Note that we check for both the expected editBox and
	-- addHistory values to ignore calls for other edit boxes (e.g., the macro
	-- execution editBox).
--	if (_debugFrame) then
--		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChatEdit_SendText("..editBox:GetName()..", "..tostring(addHistory or "")..")");
--	end
	if ((editBox == ChatFrameEditBox) and (addHistory == 1)) then
		local type = editBox:GetAttribute("chatType");
		if (type == "CHANNEL") then
			local channelTarget = editBox:GetAttribute("channelTarget");
			local index = tonumber(channelTarget);
			local cdata = _channelList[index or 0];
			local sticky = nil;
			if (cdata and (cdata.isSticky == 1)) then
				if (ChatmanagerState.Options.SmartSticky == 1) then
					if ((GetNumRaidMembers() == 0) and (GetNumPartyMembers() == 0)) then
						sticky = 1;
					end
				else
					sticky = 1;
				end
			end
			if (sticky == 1) then
				editBox:SetAttribute("stickyType", type);
				editBox.stickyChannel = channelTarget;
			elseif ((editBox:GetAttribute("stickyType") == "CHANNEL") and editBox.stickyChannel) then
				editBox:SetAttribute("channelTarget", editBox.stickyChannel);
			end
		elseif (type and ChatTypeInfo[type] and (ChatTypeInfo[type].sticky == 1)) then
			editBox.stickyChannel = nil;
		end
	end
end


function Chatmanager.UnitPopup_ShowMenu(dropdownMenu, which, unit, name, userData)
	-- Called via hooksecurefunc after UnitPopup_ShowMenu has been processed,
	-- this function allows us to modify the name (title) of the displayed menu
	-- if needed without preventing the menu from operating correctly.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.UnitPopup_ShowMenu("..tostring(dropdownMenu and dropdownMenu.GetName and dropdownMenu:GetName())..", "..tostring(which)..", "..tostring(unit)..", "..tostring(name)..", "..tostring(userData)..")");
	end
	if ((dropdownMenu == FriendsDropDown) and (which == "FRIEND") and name and (ChatmanagerState.Options.ShowPlayerNote == 1)) then
		local note = Chatmanager.GetPlayerOrGuildNote(name);
		if (note and (note ~= "")) then
			local button = DropDownList1Button1;
			local normalText = DropDownList1Button1NormalText;
			if (normalText:GetText() == name) then
				local oldWidth = normalText:GetWidth();
				for i = 2, DropDownList1.numButtons do
					local width = getglobal("DropDownList1Button"..i.."NormalText"):GetWidth();
					if (width > oldWidth) then
						oldWidth = width;
					end
				end
				button:SetText(string.format(CHATMANAGER_TEXT.CHANNEL_JOIN_NOTE_FORMAT, name, note));
				local newWidth = normalText:GetWidth();
				if (newWidth > oldWidth) then
					DropDownList1.maxWidth = DropDownList1.maxWidth + newWidth - oldWidth;
					DropDownList1:SetWidth(DropDownList1.maxWidth + 25);
				end
			end
		end
	end
end


function Chatmanager.ChatFrame_OpenChat(text, chatFrame)
	-- We hook ChatFrame_OpenChat so that we can "correct" the default UI's
	-- tendency to drop you back into say anytime you leave a group, etc.
	if (not chatFrame) then
		chatFrame = ChatFrame1;
	end
	local stickyType = chatFrame.editBox:GetAttribute("stickyType");
	local chatType = chatFrame.editBox:GetAttribute("chatType");
	if (stickyType and (chatType == "SAY") and (chatType ~= stickyType)) then
		if ((stickyType == "RAID") and (ChatTypeInfo[stickyType].sticky == 1) and (GetNumRaidMembers() > 0)) then
			chatFrame.editBox:SetAttribute("chatType", "RAID");
			ChatEdit_UpdateHeader(chatFrame.editBox);
		elseif ((stickyType == "BATTLEGROUND") and (ChatTypeInfo[stickyType].sticky == 1) and (GetNumRaidMembers() > 0)) then
			chatFrame.editBox:SetAttribute("chatType", "BATTLEGROUND");
			ChatEdit_UpdateHeader(chatFrame.editBox);
		elseif ((stickyType == "PARTY") and (ChatTypeInfo[stickyType].sticky == 1) and (GetNumPartyMembers() > 0)) then
			chatFrame.editBox:SetAttribute("chatType", "PARTY");
			ChatEdit_UpdateHeader(chatFrame.editBox);
		elseif ((stickyType == "CHANNEL") and editBox.stickyChannel) then
			editBox:SetAttribute("channelTarget", editBox.stickyChannel);
			chatFrame.editBox:SetAttribute("chatType", "CHANNEL");
			ChatEdit_UpdateHeader(chatFrame.editBox);
		elseif ((ChatTypeInfo["GUILD"].sticky == 1) and (ChatTypeInfo["GUILD"].sticky == 1) and IsInGuild()) then
			chatFrame.editBox:SetAttribute("chatType", "GUILD");
			chatFrame.editBox:SetAttribute("stickyType", "GUILD");
			ChatEdit_UpdateHeader(chatFrame.editBox);
		end
	end
end


function Chatmanager.FCF_SetChatWindowFontSize(chatFrame, fontSize)
	-- Hooked so that we can have the editbox font size follow that of the
	-- ChatFrame1 text window
	if (not chatFrame) then
		chatFrame = FCF_GetCurrentChatFrame();
	end
	if (chatFrame == ChatFrame1) then
		Chatmanager.UpdateEditBoxFont(fontSize)
	end
end


function Chatmanager.ChannelList_Update()
	-- Called after the original ChannelList_Update function, this iterates
	-- through the buttons to assign them the correct IDs etc. from a
	-- Chatmanager perspective.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChannelList_Update()");
	end
	if (not ChannelFrame:IsShown()) then
		return;
	end

	-- Set flag to avoid confusion in ChannelList_UpdateVoice
	_channelListUpdateInProgress = 1;

	-- Build some maps of the buttons and their data 
	local channelWorldButtons = {};
	local channelWorldInfo = {};
	local channelCustomButtons = {};
	local channelCustomInfo = {};
	local selectedID = GetSelectedDisplayChannel() or (GetActiveVoiceChannel and GetActiveVoiceChannel());
	for i = 1, GetNumDisplayChannels() do
		local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo(i);
		if (header) then
			if (count and ChatmanagerFrame:IsShown()) then
				local button = getglobal("ChannelButton"..i);
				local buttonCollapsed = getglobal("ChannelButton"..i.."Collapsed");
				buttonCollapsed:Hide();
				button:Disable();
			end
		elseif (name and not header and not collapsed and channelNumber and (category ~= "CHANNEL_CATEGORY_GROUP")) then
			local index = tonumber(channelNumber);
			local cdata = _channelList[index or 0];
			if (cdata and cdata.name) then
				local bdata = {};
				bdata.id = i;
				bdata.button = getglobal("ChannelButton"..i);
				bdata.buttonName = getglobal("ChannelButton"..i.."Text");
				local info = {};
				info.id = i;
				info.name = name;
				info.channelNumber = cdata.displayIndex;
				info.count = count;
				info.active = active;
				info.voiceEnabled = voiceEnabled;
				info.voiceActive = voiceActive;
				if (i == selectedID) then
					info.selected = 1;
				end
				if (category == "CHANNEL_CATEGORY_WORLD") then
					tinsert(channelWorldButtons, bdata);
					tinsert(channelWorldInfo, info);
				else
					tinsert(channelCustomButtons, bdata);
					tinsert(channelCustomInfo, info);
				end
			end
		end
	end

	-- Sort info tables into displayed channel number order (within their
	-- category).
	table.sort(channelWorldInfo, function(first, second) return first.channelNumber < second.channelNumber end);
	table.sort(channelCustomInfo, function(first, second) return first.channelNumber < second.channelNumber end);

	-- Repaint the buttons, and build the ID to button index lookup table
	_channelListButtonLookupById = {};
	for index, bdata in ipairs(channelWorldButtons) do
		local info = channelWorldInfo[index];
		if (info.active) then
			bdata.buttonName:SetText(HIGHLIGHT_FONT_COLOR_CODE..info.channelNumber..". "..info.name..FONT_COLOR_CODE_CLOSE);
			bdata.button:Enable();
			if (info.selected) then
				bdata.button:LockHighlight();
			else
				bdata.button:UnlockHighlight();
			end
		else
			bdata.buttonName:SetText(GRAY_FONT_COLOR_CODE..info.channelNumber..". "..info.name..FONT_COLOR_CODE_CLOSE);			
			bdata.button:Disable();
			bdata.button:UnlockHighlight();
		end
		if (IsVoiceChatEnabled()) then
			ChannelList_UpdateVoice(bdata.id, info.voiceEnabled, info.voiceActive);
		else
			ChannelList_UpdateVoice(bdata.id, nil, nil);
		end
		bdata.button.channel = info.name;
		bdata.button.active = info.active;
		bdata.button.channelNumber = info.channelNumber;
		bdata.button:SetID(info.id);
		_channelListButtonLookupById[info.id] = bdata.id;
	end
	for index, bdata in ipairs(channelCustomButtons) do
		local info = channelCustomInfo[index];
		if (info.active) then
			if (info.count) then
				bdata.buttonName:SetText(HIGHLIGHT_FONT_COLOR_CODE..info.channelNumber..". "..info.name.." ("..info.count..")"..FONT_COLOR_CODE_CLOSE);
			else
				bdata.buttonName:SetText(HIGHLIGHT_FONT_COLOR_CODE..info.channelNumber..". "..info.name..FONT_COLOR_CODE_CLOSE);
			end
			bdata.button:Enable();
			if (info.selected) then
				bdata.button:LockHighlight();
			else
				bdata.button:UnlockHighlight();
			end
		else
			bdata.buttonName:SetText(GRAY_FONT_COLOR_CODE..info.channelNumber..". "..info.name..FONT_COLOR_CODE_CLOSE);			
			bdata.button:Disable();
			bdata.button:UnlockHighlight();
		end
		if (IsVoiceChatEnabled()) then
			ChannelList_UpdateVoice(bdata.id, info.voiceEnabled, info.voiceActive);
		else
			ChannelList_UpdateVoice(bdata.id, nil, nil);
		end
		bdata.button.channel = info.name;
		bdata.button.active = info.active;
		bdata.button.channelNumber = info.channelNumber;
		bdata.button:SetID(info.id);
		_channelListButtonLookupById[info.id] = bdata.id;
	end

	-- Clear flag
	_channelListUpdateInProgress = nil;
end


function Chatmanager.ChannelList_CountUpdate(id, count)
	-- The behaviour of this function in the Blizzard code has changed from
	-- earlier versions, and now only shows counts for groups (party and
	-- raid), so we don't need to worry about the count being applied to an
	-- incorrect channel.
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChannelList_CountUpdate("..tostring(id)..", "..tostring(count)..")");
	end
	local name, header, _, channelNumber, _, active = GetChannelDisplayInfo(id);
	if (name and not header and active and count and (count > 0)) then
		local index = tonumber(channelNumber);
		local cdata = _channelList[index or 0];
		if (cdata and cdata.name) then
			cdata.playerCount = count;
			Chatmanager.Frame_UpdateChannelPlayerCounts(cdata);
			if (index == _selectedChannelIndex) then
				Chatmanager.Frame_UpdatePlayerList();
			end
		end
	end
end


function Chatmanager.ChannelList_UpdateHighlight(id)
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.ChannelList_UpdateHighlight("..tostring(id)..")");
	end
	if (not id or not ChannelFrame:IsShown()) then
		return;
	end
	for i = 1, GetNumDisplayChannels() do
		button = getglobal("ChannelButton"..i);
		if (button:GetID() == id) then
			button:LockHighlight();
			local displayIndex = tonumber(button.channelNumber);
			if (displayIndex) then
				if (displayIndex ~= _selectedChannelDisplayIndex) then
					Chatmanager.Frame_UpdateChannelList(displayIndex);
					Chatmanager.Frame_UpdatePlayerList();
				end
			elseif (_selectedChannelDisplayIndex) then
				Chatmanager.Frame_UpdateChannelList("");
				Chatmanager.Frame_UpdatePlayerList();
			end
		else
			button:UnlockHighlight();
		end
	end
end


function Chatmanager.CreateChatChannelList(...)
	-- We need to change the numbers shown for each channel (which fortunately
	-- the main code will then sort into the correct order for us ;).
	-- Although modifying the global will taint it, it doesn't seem to impact
	-- anything important.  Called by posthook...
	if ( not FCF_GetCurrentChatFrame() ) then
		return;
	end
	for _, ldata in ipairs(CHAT_CONFIG_CHANNEL_LIST) do
		local name = ldata.channelName;
		local lname = string.lower(name);
		local index = _channelIndexLookupByName[lname];
		if (index and _channelList[index] and (_channelList[index].name == name)) then
			ldata.displayIndex = _channelList[index].displayIndex;
			ldata.text = ldata.displayIndex.."."..name;
		end
	end
end


------------------------------------------------------------------------------
-- Hooked functions from other addons
------------------------------------------------------------------------------
function Chatmanager.FLT_ParseChatMessage(text)
	-- If we have already converted any links in the text, make sure to not
	-- have Linkerator convert them back again!
	if (_ignoreLinkConversions) then
		return text;
	else
		return _original_FLT_ParseChatMessage(text);
	end
end


------------------------------------------------------------------------------
-- Exported access functions
------------------------------------------------------------------------------
function Chatmanager.GetChannelInfo(channelNameOrNum)
	-- Given a channel name or number, this function returns the display index,
	-- channel number, short channel name, full channel name, and the channel
	-- type (0 for private).
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.GetChannelInfo("..tostring(channelNameOrNum or "nil")..")");
	end
	local index = tonumber(channelNameOrNum);
	if (not index) then
		local lname = string.lower(channelNameOrNum);
		index = _channelIndexLookupByName[lname] or 0;
	end
	local cdata = _channelList[index or 0];
	if (cdata) then
		return cdata.displayIndex, index, cdata.name, cdata.fullName, cdata.type;
	end
end


------------------------------------------------------------------------------
-- Chat frame OnScroll and OnShow handlers
------------------------------------------------------------------------------
function Chatmanager.ChatFrameOnScroll(self)
	-- Use the same key combinations and scrolling amounts as ChatBox (shift
	-- for top/bottom, control for by 3 lines, or normally by 1 line).
	if (arg1 == 1) then
		if (IsShiftKeyDown()) then
			self:ScrollToTop();
		elseif (IsControlKeyDown()) then
			self:ScrollUp();
			self:ScrollUp();
			self:ScrollUp();
		else
			self:ScrollUp();
		end
	elseif (arg1 == -1) then
		if (IsShiftKeyDown()) then
			self:ScrollToBottom();
		elseif (IsControlKeyDown()) then
			self:ScrollDown();
			self:ScrollDown();
			self:ScrollDown();
		else
			self:ScrollDown();
		end
	end
	if (_hideChatButtons == 1) then
		if (self:AtBottom()) then
			getglobal(self:GetName().."BottomButton"):Hide();
		else
			getglobal(self:GetName().."BottomButton"):Show();
		end
	end
end


function Chatmanager.ChatFrameOnShow(self)
	if (self.chatmanagerOrigOnShow) then
		self.chatmanagerOrigOnShow(self);
		if (self:AtBottom()) then
			getglobal(self:GetName().."BottomButton"):Hide();
		end
		getglobal(self:GetName().."DownButton"):Hide();
		getglobal(self:GetName().."UpButton"):Hide();
	end
end


function Chatmanager.ChatFrameBottomButtonOnClick(self, button, down)
	if (self.chatmanagerOrigOnClick) then
		self.chatmanagerOrigOnClick(self, button, down);
		self:Hide();
	end
end


------------------------------------------------------------------------------
-- OnEvent handler
------------------------------------------------------------------------------
function Chatmanager.OnEvent(event, arg1)
	if (_debugFrame) then
		_debugFrame:AddMessage(CHATMANAGER_TEXT.DEBUG.."Chatmanager.OnEvent("..event..", "..(arg1 or "")..")");
	end
	if (event == "PLAYER_TARGET_CHANGED") then
		if (UnitIsPlayer("target")) then
			local name = UnitName("target");
			if (_selectedPlayer ~= name) then
				if (name and (name ~= "")) then
					_selectedPlayer = name;
				else
					_selectedPlayer = nil;
				end
				Chatmanager.Frame_UpdatePlayerList();
			end
		elseif (_selectedPlayer) then
			_selectedPlayer = nil;
			Chatmanager.Frame_UpdatePlayerList();
		end

	elseif (event == "CHANNEL_UI_UPDATE") then
		Chatmanager.UpdateChannelList();

	elseif (event == "CHANNEL_ROSTER_UPDATE") then
		Chatmanager.UpdateChannelRoster(arg1);

	elseif (event == "CHANNEL_PASSWORD_REQUEST") then
		Chatmanager.ProcessPasswordRequest(arg1);

	elseif (event == "PLAYER_REGEN_DISABLED") then
		Chatmanager.Frame_HideSecureButtons();

	elseif (event == "PLAYER_REGEN_ENABLED") then
		Chatmanager.Frame_ShowSecureButtons();

	elseif (event == "PLAYER_LEAVING_WORLD") then
		_selectedPlayer = nil;
		Chatmanager.UnregisterEvents();
		_thisFrame:SetScript("OnUpdate", nil);

	elseif (event == "PLAYER_ENTERING_WORLD") then
		Chatmanager.RegisterEvents();
		Chatmanager.Frame_UpdatePlayerList();
		if (_playerListUpdateTimeout or _channelConnects) then
			_thisFrame:SetScript("OnUpdate", Chatmanager.Frame_OnUpdate);
		end

	elseif (event == "VARIABLES_LOADED") then
		_serverName = GetRealmName();
		Chatmanager.VariablesLoaded();
		if (_serverName and _playerName) then
			Chatmanager.PlayerLogin();
		end

	elseif (event == "PLAYER_LOGIN") then
		_playerName = UnitName("player");
		if (_serverName and _playerName) then
			Chatmanager.PlayerLogin();
		end

	elseif (event == "PLAYER_LOGOUT") then
		if (_serverName and _playerName) then
			Chatmanager.PlayerLogout();
		end

	end
end


------------------------------------------------------------------------------
-- OnLoad function
------------------------------------------------------------------------------
function Chatmanager.OnLoad(self)
	-- Record our frame pointer for later
	_thisFrame = self;

	-- Register for player events
	_thisFrame:RegisterEvent("VARIABLES_LOADED");
	_thisFrame:RegisterEvent("PLAYER_LOGIN");
	_thisFrame:RegisterEvent("PLAYER_LOGOUT");
	_thisFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
	_thisFrame:RegisterEvent("PLAYER_LEAVING_WORLD");

	-- Register slash functions
	SLASH_CHATMANAGER1 = "/chatmanager";
	SLASH_CHATMANAGER2 = "/chatman";
	SlashCmdList["CHATMANAGER"] = function(text)
		Chatmanager.SlashCommand(text);
	end
	SLASH_GROUP1 = "/group";
	SLASH_GROUP2 = "/gr";
	SlashCmdList["GROUP"] = function(text)
		Chatmanager.Group_SlashCommand(text);
	end

	-- Add our static popup dialogs for the actions we need
	StaticPopupDialogs["CHATMANAGER_LEAVE_CONFIRM"] = _chatmanagerConfirmLeavePopup;
	StaticPopupDialogs["CHATMANAGER_NEW_CHANNEL"] = _chatmanagerEnterChannelNamePopup;
	StaticPopupDialogs["CHATMANAGER_PLAYER_NOTE"] = _chatmanagerEnterPlayerNotePopup;

	-- Generate the search text for _playerIsOnlineSearchPattern
	local temp = string.gsub(ERR_FRIEND_ONLINE_SS, "([%^%$%(%)%.%[%]%*%+%-%?])", "%%%1");
	_playerIsOnlineSearchPattern = "^"..string.gsub(temp, "%%s", "(.+)");

	-- Register with Blizzard interface options
	Chatmanager.RegisterOptions(CHATMANAGER_NAME, CHATMANAGER_NAME.." v"..CHATMANAGER_VERSION, CHATMANAGER_DESCRIPTION, CHATMANAGER_HELP);
end


------------------------------------------------------------------------------
-- Slash command function
------------------------------------------------------------------------------
function Chatmanager.SlashCommand(text)
	if (text) then
		local command, params = Chatmanager.GetNextParam(text);
		command = string.lower(command or "");
		if (command == CHATMANAGER_COMMANDS.COMMAND_SHOW) then
			ShowUIPanel(ChatmanagerFrame);

		elseif (command == CHATMANAGER_COMMANDS.COMMAND_HIDE) then
			HideUIPanel(ChatmanagerFrame);
			HideUIPanel(ChatmanagerOptionsFrame);

		elseif (command == CHATMANAGER_COMMANDS.COMMAND_OPTIONS) then
			ShowUIPanel(ChatmanagerOptionsFrame);

		elseif (command == CHATMANAGER_COMMANDS.COMMAND_RESET) then
			params = string.lower(params or "");
			if (params == "") then
				DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_RESET_HELP);
			end
			if ((params == CHATMANAGER_COMMANDS.OPTION_RESET_ORDER) or (params == CHATMANAGER_COMMANDS.OPTION_RESET_ALL)) then
				Chatmanager.ResetOrdering();
				DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_RESET_ORDER_CONFIRM);
			end
			if ((params == CHATMANAGER_COMMANDS.OPTION_RESET_COLOR) or (params == CHATMANAGER_COMMANDS.OPTION_RESET_ALL)) then
				Chatmanager.ResetColors();
				DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_RESET_COLOR_CONFIRM);
			end
			if ((params == CHATMANAGER_COMMANDS.OPTION_RESET_WINDOW) or (params == CHATMANAGER_COMMANDS.OPTION_RESET_ALL)) then
				Chatmanager.ResetWindows();
				DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_RESET_WINDOW_CONFIRM);
			end
			Chatmanager.Frame_OnShow();

		elseif ((command == CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE) or (command == CHATMANAGER_COMMANDS.COMMAND_PLAYEREDIT)) then
			local pname, pnote = Chatmanager.GetNextParam(params or "");
			if (pname ~= "") then
				pname = string.upper(string.sub(pname, 1, 1))..string.lower(string.sub(pname, 2));
				if (pnote and (pnote ~= "") and (string.gsub(pnote, "%s", "") ~= "")) then
					pnote = string.sub(pnote, 1, _chatmanagerEnterPlayerNotePopup.maxLetters);
					ChatmanagerState.Servers[_serverName].Notes[pname] = pnote;
					Chatmanager.Frame_UpdatePlayerList();
					DEFAULT_CHAT_FRAME:AddMessage(string.format(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_CONFIRM_FORMAT, pname, pnote));
				else
					CloseDropDownMenus();
					ColorPickerFrame:Hide();
					ChatmanagerDropDown.thisName = pname;
					local dialogFrame = StaticPopup_Show("CHATMANAGER_PLAYER_NOTE", ChatmanagerDropDown.thisName);
					if (dialogFrame) then
						dialogFrame.data = ChatmanagerDropDown.thisName;
					end
				end
			else
				DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_HELP);
			end

		elseif (command == CHATMANAGER_COMMANDS.COMMAND_LISTNOTES) then
			local found = 0;
			for pname, pnote in pairs(ChatmanagerState.Servers[_serverName].Notes) do
				found = found + 1;
				DEFAULT_CHAT_FRAME:AddMessage(string.format(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_LIST_FORMAT, found, pname, pnote));
			end
			if (found < 1) then
				DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_PLAYERNOTE_LIST_EMPTY);
			end

		elseif (command == CHATMANAGER_COMMANDS.COMMAND_DEBUGON) then
			_debugFrame = nil;
			local frameNum = tonumber(params or "");
			if (frameNum and (frameNum >= 1) and (frameNum <= 7)) then
				_debugFrame = getglobal("ChatFrame"..frameNum);
			end
			if (not _debugFrame) then
				_debugFrame = DEFAULT_CHAT_FRAME;
			end
			DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_DEBUGON_CONFIRM);

		elseif (command == CHATMANAGER_COMMANDS.COMMAND_DEBUGOFF) then
			_debugFrame = nil;
			DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_COMMANDS.COMMAND_DEBUGOFF_CONFIRM);

		elseif (command == "") then
			if (ChatmanagerFrame:IsShown() or ChatmanagerOptionsFrame:IsShown()) then
				HideUIPanel(ChatmanagerFrame);
				HideUIPanel(ChatmanagerOptionsFrame);
			else
				ShowUIPanel(ChatmanagerFrame);
			end

		else
			DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_EM.ON..CHATMANAGER_NAME.." v"..CHATMANAGER_VERSION..CHATMANAGER_EM.OFF);
			DEFAULT_CHAT_FRAME:AddMessage(CHATMANAGER_DESCRIPTION);
			for _, string in ipairs(CHATMANAGER_HELP) do
				DEFAULT_CHAT_FRAME:AddMessage(string);
			end
		end
	end
end


function Chatmanager.Group_SlashCommand(text)
	-- Sends a message to RAID if in a raid, PARTY if in a party, or SAY
	-- otherwise
	if (text ~= "") then
		if (GetNumRaidMembers() > 0) then
			SendChatMessage(text, "RAID", nil);
		elseif (GetNumPartyMembers() > 0) then
			SendChatMessage(text, "PARTY", nil);
		else
			SendChatMessage(text, "SAY", nil);
		end
	end
end


------------------------------------------------------------------------------
-- Whoami functions
------------------------------------------------------------------------------
function Chatmanager:GetName()
	return "Chatmanager";
end


function Chatmanager:GetTitle()
	return CHATMANAGER_NAME;
end


function Chatmanager:GetVersion()
	return CHATMANAGER_VERSION;
end


------------------------------------------------------------------------------
-- Blizzard options panel functions
------------------------------------------------------------------------------
function Chatmanager.RegisterOptions(name, titleText, descriptionText, helpText)
	local panel = CreateFrame("Frame", nil);
	panel.name = name;

	local title = panel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge");
	title:SetPoint("TOPLEFT", panel, "TOPLEFT", 15, -15);
	title:SetPoint("TOPRIGHT", panel, "TOPRIGHT", -15, -15);
	title:SetJustifyH("LEFT");
	title:SetJustifyV("TOP");
	title:SetText(titleText);
	local last = title;
	local spacing = 10;

	if (descriptionText) then
		local description = panel:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall");
		description:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -3);
		description:SetWidth(380);
		description:SetJustifyH("LEFT");
		description:SetJustifyV("TOP");
		description:SetNonSpaceWrap(1);
		description:SetText(descriptionText);
		last = description;
	end

	if (helpText) then
		local helpTextList = helpText;
		if (type(helpText) == "string") then
			helpTextList = {helpText};
		end
		for _, text in ipairs(helpTextList) do
			local line = panel:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall");
			line:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -(1 + spacing));
			line:SetWidth(380);
			line:SetJustifyH("LEFT");
			line:SetJustifyV("TOP");
			line:SetNonSpaceWrap(1);
			local uncolored = string.gsub(text, "(|c%x%x%x%x%x%x%x%x)", "");
			if (string.sub(uncolored, 1, 1) == "/") then
				line:SetMultilineIndent(16);
			else
				line:SetMultilineIndent(0);
			end
			line:SetSpacing(1);
			if (text ~= "") then
				line:SetText(text);
			else
				line:SetText(" ");
			end
			last = line;
			spacing = 0;
		end
	end

	InterfaceOptions_AddCategory(panel);
end

